import boto3

from moto import mock_aws
from tests import DEFAULT_ACCOUNT_ID


@mock_aws
def test_provision_permission_set():
    ssoadmin = boto3.client("sso-admin", "us-east-1")

    instance_arn = ssoadmin.list_instances()["Instances"][0]["InstanceArn"]

    p_set_arn = ssoadmin.create_permission_set(InstanceArn=instance_arn, Name="pset1")[
        "PermissionSet"
    ]["PermissionSetArn"]

    status = ssoadmin.provision_permission_set(
        InstanceArn=instance_arn,
        PermissionSetArn=p_set_arn,
        TargetType="AWS_ACCOUNT",
    )["PermissionSetProvisioningStatus"]

    assert status["AccountId"] == DEFAULT_ACCOUNT_ID
    assert status["CreatedDate"]
    assert status["PermissionSetArn"] == p_set_arn
    assert status["Status"] == "SUCCEEDED"


@mock_aws
def test_list_permission_sets_provisioned_to_account():
    ssoadmin = boto3.client("sso-admin", "us-east-1")

    instance_arn = ssoadmin.list_instances()["Instances"][0]["InstanceArn"]

    p_set_arn = ssoadmin.create_permission_set(InstanceArn=instance_arn, Name="pset1")[
        "PermissionSet"
    ]["PermissionSetArn"]

    provisioned = ssoadmin.list_permission_sets_provisioned_to_account(
        AccountId=DEFAULT_ACCOUNT_ID, InstanceArn=instance_arn
    )["PermissionSets"]
    assert len(provisioned) == 0

    accounts = ssoadmin.list_accounts_for_provisioned_permission_set(
        InstanceArn=instance_arn, PermissionSetArn=p_set_arn
    )["AccountIds"]
    assert accounts == []

    ssoadmin.provision_permission_set(
        InstanceArn=instance_arn,
        PermissionSetArn=p_set_arn,
        TargetType="AWS_ACCOUNT",
    )

    provisioned = ssoadmin.list_permission_sets_provisioned_to_account(
        AccountId=DEFAULT_ACCOUNT_ID, InstanceArn=instance_arn
    )["PermissionSets"]
    assert provisioned == [p_set_arn]

    accounts = ssoadmin.list_accounts_for_provisioned_permission_set(
        InstanceArn=instance_arn, PermissionSetArn=p_set_arn
    )["AccountIds"]
    assert accounts == [DEFAULT_ACCOUNT_ID]



from typing import Optional

from moto.stepfunctions.parser.api import HistoryEventType, MapStateStartedEventDetails
from moto.stepfunctions.parser.asl.component.common.catch.catch_decl import CatchDecl
from moto.stepfunctions.parser.asl.component.common.catch.catch_outcome import (
    CatchOutcome,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.common.parameters import Parameters
from moto.stepfunctions.parser.asl.component.common.path.items_path import ItemsPath
from moto.stepfunctions.parser.asl.component.common.path.result_path import ResultPath
from moto.stepfunctions.parser.asl.component.common.result_selector import (
    ResultSelector,
)
from moto.stepfunctions.parser.asl.component.common.retry.retry_decl import RetryDecl
from moto.stepfunctions.parser.asl.component.common.retry.retry_outcome import (
    RetryOutcome,
)
from moto.stepfunctions.parser.asl.component.state.exec.execute_state import (
    ExecutionState,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.item_reader_decl import (
    ItemReader,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_selector import (
    ItemSelector,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.distributed_item_processor import (
    DistributedItemProcessor,
    DistributedItemProcessorEvalInput,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.inline_item_processor import (
    InlineItemProcessor,
    InlineItemProcessorEvalInput,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.item_processor_decl import (
    ItemProcessorDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.item_processor_factory import (
    from_item_processor_decl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_component import (
    IterationComponent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.iterator import (
    Iterator,
    IteratorEvalInput,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.iterator_decl import (
    IteratorDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.max_concurrency import (
    MaxConcurrency,
)
from moto.stepfunctions.parser.asl.component.state.state_props import StateProps
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails


class StateMap(ExecutionState):
    items_path: ItemsPath
    iteration_component: IterationComponent
    item_reader: Optional[ItemReader]
    item_selector: Optional[ItemSelector]
    parameters: Optional[Parameters]
    max_concurrency: MaxConcurrency
    result_path: Optional[ResultPath]
    result_selector: ResultSelector
    retry: Optional[RetryDecl]
    catch: Optional[CatchDecl]

    def __init__(self):
        super(StateMap, self).__init__(
            state_entered_event_type=HistoryEventType.MapStateEntered,
            state_exited_event_type=HistoryEventType.MapStateExited,
        )

    def from_state_props(self, state_props: StateProps) -> None:
        super(StateMap, self).from_state_props(state_props)
        self.items_path = state_props.get(ItemsPath) or ItemsPath()
        self.item_reader = state_props.get(ItemReader)
        self.item_selector = state_props.get(ItemSelector)
        self.parameters = state_props.get(Parameters)
        self.max_concurrency = state_props.get(MaxConcurrency) or MaxConcurrency()
        self.result_path = state_props.get(ResultPath) or ResultPath(
            result_path_src=ResultPath.DEFAULT_PATH
        )
        self.result_selector = state_props.get(ResultSelector)
        self.retry = state_props.get(RetryDecl)
        self.catch = state_props.get(CatchDecl)

        # TODO: add check for itemreader used in distributed mode only.

        iterator_decl = state_props.get(typ=IteratorDecl)
        item_processor_decl = state_props.get(typ=ItemProcessorDecl)

        if iterator_decl and item_processor_decl:
            raise ValueError("Cannot define both Iterator and ItemProcessor.")

        iteration_decl = iterator_decl or item_processor_decl
        if iteration_decl is None:
            raise ValueError(
                f"Missing ItemProcessor/Iterator definition in props '{state_props}'."
            )

        if isinstance(iteration_decl, IteratorDecl):
            self.iteration_component = Iterator.from_declaration(iteration_decl)
        elif isinstance(iteration_decl, ItemProcessorDecl):
            self.iteration_component = from_item_processor_decl(iteration_decl)
        else:
            raise ValueError(f"Unknown value for IteratorDecl '{iteration_decl}'.")

    def _eval_execution(self, env: Environment) -> None:
        self.items_path.eval(env)
        if self.item_reader:
            env.event_history.add_event(
                context=env.event_history_context,
                hist_type_event=HistoryEventType.MapStateStarted,
                event_detail=EventDetails(
                    mapStateStartedEventDetails=MapStateStartedEventDetails(length=0)
                ),
            )
            input_items = None
        else:
            input_items = env.stack.pop()
            env.event_history.add_event(
                context=env.event_history_context,
                hist_type_event=HistoryEventType.MapStateStarted,
                event_detail=EventDetails(
                    mapStateStartedEventDetails=MapStateStartedEventDetails(
                        length=len(input_items)
                    )
                ),
            )

        if isinstance(self.iteration_component, Iterator):
            eval_input = IteratorEvalInput(
                state_name=self.name,
                max_concurrency=self.max_concurrency.num,
                input_items=input_items,
                parameters=self.parameters,
            )
        elif isinstance(self.iteration_component, InlineItemProcessor):
            eval_input = InlineItemProcessorEvalInput(
                state_name=self.name,
                max_concurrency=self.max_concurrency.num,
                input_items=input_items,
                item_selector=self.item_selector,
            )
        elif isinstance(self.iteration_component, DistributedItemProcessor):
            eval_input = DistributedItemProcessorEvalInput(
                state_name=self.name,
                max_concurrency=self.max_concurrency.num,
                item_reader=self.item_reader,
                item_selector=self.item_selector,
            )
        else:
            raise RuntimeError(
                f"Unknown iteration component of type '{type(self.iteration_component)}' '{self.iteration_component}'."
            )

        env.stack.append(eval_input)
        self.iteration_component.eval(env)

        env.event_history.add_event(
            context=env.event_history_context,
            hist_type_event=HistoryEventType.MapStateSucceeded,
            update_source_event_id=False,
        )

    def _eval_state(self, env: Environment) -> None:
        # Initialise the retry counter for execution states.
        env.context_object_manager.context_object["State"]["RetryCount"] = 0

        # Attempt to evaluate the state's logic through until it's successful, caught, or retries have run out.
        while True:
            try:
                self._evaluate_with_timeout(env)
                break
            except Exception as ex:
                failure_event: FailureEvent = self._from_error(env=env, ex=ex)

                if self.retry:
                    retry_outcome: RetryOutcome = self._handle_retry(
                        env=env, failure_event=failure_event
                    )
                    if retry_outcome == RetryOutcome.CanRetry:
                        continue

                env.event_history.add_event(
                    context=env.event_history_context,
                    hist_type_event=HistoryEventType.MapStateFailed,
                )

                if self.catch:
                    catch_outcome: CatchOutcome = self._handle_catch(
                        env=env, failure_event=failure_event
                    )
                    if catch_outcome == CatchOutcome.Caught:
                        break

                self._handle_uncaught(env=env, failure_event=failure_event)



import datetime
import json

import boto3

from moto import mock_aws


@mock_aws
def test_put_telemetry():
    client = boto3.client("xray", region_name="us-east-1")

    client.put_telemetry_records(
        TelemetryRecords=[
            {
                "Timestamp": datetime.datetime(2015, 1, 1),
                "SegmentsReceivedCount": 123,
                "SegmentsSentCount": 123,
                "SegmentsSpilloverCount": 123,
                "SegmentsRejectedCount": 123,
                "BackendConnectionErrors": {
                    "TimeoutCount": 123,
                    "ConnectionRefusedCount": 123,
                    "HTTPCode4XXCount": 123,
                    "HTTPCode5XXCount": 123,
                    "UnknownHostCount": 123,
                    "OtherCount": 123,
                },
            }
        ],
        EC2InstanceId="string",
        Hostname="string",
        ResourceARN="string",
    )


@mock_aws
def test_put_trace_segments():
    client = boto3.client("xray", region_name="us-east-1")

    client.put_trace_segments(
        TraceSegmentDocuments=[
            json.dumps(
                {
                    "name": "example.com",
                    "id": "70de5b6f19ff9a0a",
                    "start_time": 1.478293361271e9,
                    "trace_id": "1-581cf771-a006649127e371903a2de979",
                    "end_time": 1.478293361449e9,
                }
            )
        ]
    )


@mock_aws
def test_trace_summary():
    client = boto3.client("xray", region_name="us-east-1")

    client.put_trace_segments(
        TraceSegmentDocuments=[
            json.dumps(
                {
                    "name": "example.com",
                    "id": "70de5b6f19ff9a0a",
                    "start_time": 1.478293361271e9,
                    "trace_id": "1-581cf771-a006649127e371903a2de979",
                    "in_progress": True,
                }
            ),
            json.dumps(
                {
                    "name": "example.com",
                    "id": "70de5b6f19ff9a0b",
                    "start_time": 1478293365,
                    "trace_id": "1-581cf771-a006649127e371903a2de979",
                    "end_time": 1478293385,
                }
            ),
        ]
    )

    client.get_trace_summaries(
        StartTime=datetime.datetime(2014, 1, 1), EndTime=datetime.datetime(2017, 1, 1)
    )


@mock_aws
def test_batch_get_trace():
    client = boto3.client("xray", region_name="us-east-1")

    client.put_trace_segments(
        TraceSegmentDocuments=[
            json.dumps(
                {
                    "name": "example.com",
                    "id": "70de5b6f19ff9a0a",
                    "start_time": 1.478293361271e9,
                    "trace_id": "1-581cf771-a006649127e371903a2de979",
                    "in_progress": True,
                }
            ),
            json.dumps(
                {
                    "name": "example.com",
                    "id": "70de5b6f19ff9a0b",
                    "start_time": 1478293365,
                    "trace_id": "1-581cf771-a006649127e371903a2de979",
                    "end_time": 1478293385,
                }
            ),
        ]
    )

    resp = client.batch_get_traces(
        TraceIds=[
            "1-581cf771-a006649127e371903a2de979",
            "1-581cf772-b006649127e371903a2de979",
        ]
    )
    assert len(resp["UnprocessedTraceIds"]) == 1
    assert len(resp["Traces"]) == 1


# Following are not implemented, just testing it returns what boto expects
@mock_aws
def test_batch_get_service_graph():
    client = boto3.client("xray", region_name="us-east-1")

    client.get_service_graph(
        StartTime=datetime.datetime(2014, 1, 1), EndTime=datetime.datetime(2017, 1, 1)
    )


@mock_aws
def test_batch_get_trace_graph():
    client = boto3.client("xray", region_name="us-east-1")

    client.batch_get_traces(
        TraceIds=[
            "1-581cf771-a006649127e371903a2de979",
            "1-581cf772-b006649127e371903a2de979",
        ]
    )



import datetime

import boto3
import pytest
from botocore.client import ClientError

from moto import mock_aws
from moto.s3.responses import DEFAULT_REGION_NAME

from . import s3_aws_verified


@pytest.mark.parametrize(
    "key_name",
    [
        "the-key",
        "the-unicode-💩-key",
        "key-with?question-mark",
        "key-with%2Fembedded%2Furl%2Fencoding",
    ],
)
@mock_aws
def test_copy_key_boto3(key_name):
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_resource.create_bucket(Bucket="foobar")

    key = s3_resource.Object("foobar", key_name)
    key.put(Body=b"some value")

    key2 = s3_resource.Object("foobar", "new-key")
    key2.copy_from(CopySource=f"foobar/{key_name}")

    resp = client.get_object(Bucket="foobar", Key=key_name)
    assert resp["Body"].read() == b"some value"
    resp = client.get_object(Bucket="foobar", Key="new-key")
    assert resp["Body"].read() == b"some value"


@pytest.mark.aws_verified
@s3_aws_verified
def test_copy_key_boto3_with_args(bucket_name=None):
    # Setup
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    key_name = "key"
    new_key = "new_key"
    expected_hash = "qz0H8xacy9DtbEtF3iFRn5+TjHLSQSSZiquUnOg7tRs="

    key = s3_resource.Object(bucket_name, key_name)
    key.put(Body=b"some value")

    # Execute
    key2 = s3_resource.Object(bucket_name, new_key)
    key2.copy(
        CopySource={"Bucket": bucket_name, "Key": key_name},
        ExtraArgs={
            "ChecksumAlgorithm": "SHA256",
            "WebsiteRedirectLocation": "http://getmoto.org/",
        },
    )

    # Verify
    resp = client.get_object_attributes(
        Bucket=bucket_name, Key=new_key, ObjectAttributes=["Checksum"]
    )

    assert "Checksum" in resp
    assert "ChecksumSHA256" in resp["Checksum"]
    assert resp["Checksum"]["ChecksumSHA256"] == expected_hash

    obj = client.get_object(Bucket=bucket_name, Key=new_key)
    assert obj["WebsiteRedirectLocation"] == "http://getmoto.org/"

    # Verify in place
    copy_in_place = client.copy_object(
        Bucket=bucket_name,
        CopySource=f"{bucket_name}/{new_key}",
        Key=new_key,
        ChecksumAlgorithm="SHA256",
        MetadataDirective="REPLACE",
    )

    assert "ChecksumSHA256" in copy_in_place["CopyObjectResult"]
    assert copy_in_place["CopyObjectResult"]["ChecksumSHA256"] == expected_hash


@pytest.mark.aws_verified
@s3_aws_verified
def test_copy_key_boto3_with_args__using_multipart(bucket_name=None):
    # Setup
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    key_name = "key"
    new_key = "new_key"
    expected_hash = "DnKotDi4EtYGwNMDKmnR6SqH3bWVOlo2BC+tsz9rHqw="

    key = s3_resource.Object(bucket_name, key_name)
    key.put(Body=b"some value")

    # Execute
    key2 = s3_resource.Object(bucket_name, new_key)
    key2.copy(
        CopySource={"Bucket": bucket_name, "Key": key_name},
        ExtraArgs={
            "ChecksumAlgorithm": "SHA256",
            "WebsiteRedirectLocation": "http://getmoto.org/",
        },
        Config=boto3.s3.transfer.TransferConfig(multipart_threshold=1),
    )

    # Verify
    resp = client.get_object_attributes(
        Bucket=bucket_name, Key=new_key, ObjectAttributes=["Checksum"]
    )

    assert "Checksum" in resp
    assert "ChecksumSHA256" in resp["Checksum"]
    assert resp["Checksum"]["ChecksumSHA256"] == expected_hash

    obj = client.get_object(Bucket=bucket_name, Key=new_key)
    assert obj["WebsiteRedirectLocation"] == "http://getmoto.org/"


@mock_aws
def test_copy_key_with_version_boto3():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_resource.create_bucket(Bucket="foobar")
    client.put_bucket_versioning(
        Bucket="foobar", VersioningConfiguration={"Status": "Enabled"}
    )

    key = s3_resource.Object("foobar", "the-key")
    key.put(Body=b"some value")
    key.put(Body=b"another value")

    all_versions = client.list_object_versions(Bucket="foobar", Prefix="the-key")[
        "Versions"
    ]
    old_version = [v for v in all_versions if not v["IsLatest"]][0]

    key2 = s3_resource.Object("foobar", "new-key")
    key2.copy_from(CopySource=f"foobar/the-key?versionId={old_version['VersionId']}")

    resp = client.get_object(Bucket="foobar", Key="the-key")
    assert resp["Body"].read() == b"another value"
    resp = client.get_object(Bucket="foobar", Key="new-key")
    assert resp["Body"].read() == b"some value"


@mock_aws
def test_copy_object_with_bucketkeyenabled_returns_the_value():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "test-copy-object-with-bucketkeyenabled"
    s3_resource.create_bucket(Bucket=bucket_name)

    key = s3_resource.Object(bucket_name, "the-key")
    key.put(Body=b"some value")

    key2 = s3_resource.Object(bucket_name, "new-key")
    key2.copy_from(
        CopySource=f"{bucket_name}/the-key",
        BucketKeyEnabled=True,
        ServerSideEncryption="aws:kms",
    )

    resp = client.get_object(Bucket=bucket_name, Key="the-key")
    src_headers = resp["ResponseMetadata"]["HTTPHeaders"]
    assert "x-amz-server-side-encryption" not in src_headers
    assert "x-amz-server-side-encryption-aws-kms-key-id" not in src_headers
    assert "x-amz-server-side-encryption-bucket-key-enabled" not in src_headers

    resp = client.get_object(Bucket=bucket_name, Key="new-key")
    target_headers = resp["ResponseMetadata"]["HTTPHeaders"]
    assert "x-amz-server-side-encryption" in target_headers
    # AWS will also return the KMS default key id - not yet implemented
    # assert "x-amz-server-side-encryption-aws-kms-key-id" in target_headers
    # This field is only returned if encryption is set to 'aws:kms'
    assert "x-amz-server-side-encryption-bucket-key-enabled" in target_headers
    assert (
        str(target_headers["x-amz-server-side-encryption-bucket-key-enabled"]).lower()
        == "true"
    )


@mock_aws
def test_copy_key_with_metadata():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_resource.create_bucket(Bucket="foobar")

    key = s3_resource.Object("foobar", "the-key")
    metadata = {"md": "Metadatastring"}
    content_type = "application/json"
    initial = key.put(Body=b"{}", Metadata=metadata, ContentType=content_type)

    client.copy_object(Bucket="foobar", CopySource="foobar/the-key", Key="new-key")

    resp = client.get_object(Bucket="foobar", Key="new-key")
    assert resp["Metadata"] == metadata
    assert resp["ContentType"] == content_type
    assert resp["ETag"] == initial["ETag"]


@mock_aws
def test_copy_key_replace_metadata():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_resource.create_bucket(Bucket="foobar")

    key = s3_resource.Object("foobar", "the-key")
    initial = key.put(Body=b"some value", Metadata={"md": "Metadatastring"})

    client.copy_object(
        Bucket="foobar",
        CopySource="foobar/the-key",
        Key="new-key",
        Metadata={"momd": "Mometadatastring"},
        MetadataDirective="REPLACE",
    )

    resp = client.get_object(Bucket="foobar", Key="new-key")
    assert resp["Metadata"] == {"momd": "Mometadatastring"}
    assert resp["ETag"] == initial["ETag"]


@mock_aws
def test_copy_key_without_changes_should_error():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "my_bucket"
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    key_name = "my_key"
    key = s3_resource.Object(bucket_name, key_name)

    s3_resource.create_bucket(Bucket=bucket_name)
    key.put(Body=b"some value")

    with pytest.raises(ClientError) as exc:
        client.copy_object(
            Bucket=bucket_name,
            CopySource=f"{bucket_name}/{key_name}",
            Key=key_name,
        )
        assert exc.value.response["Error"]["Message"] == (
            "This copy request is illegal because it is trying to copy an "
            "object to itself without changing the object's metadata, storage "
            "class, website redirect location or encryption attributes."
        )


@mock_aws
def test_copy_key_without_changes_should_not_error():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "my_bucket"
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    key_name = "my_key"
    key = s3_resource.Object(bucket_name, key_name)

    s3_resource.create_bucket(Bucket=bucket_name)
    key.put(Body=b"some value")

    client.copy_object(
        Bucket=bucket_name,
        CopySource=f"{bucket_name}/{key_name}",
        Key=key_name,
        Metadata={"some-key": "some-value"},
        MetadataDirective="REPLACE",
    )

    new_object = client.get_object(Bucket=bucket_name, Key=key_name)

    assert new_object["Metadata"] == {"some-key": "some-value"}


@mock_aws
def test_copy_key_reduced_redundancy():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket = s3_resource.Bucket("test_bucket")
    bucket.create()

    bucket.put_object(Key="the-key", Body=b"somedata")

    client.copy_object(
        Bucket="test_bucket",
        CopySource="test_bucket/the-key",
        Key="new-key",
        StorageClass="REDUCED_REDUNDANCY",
    )

    keys = {k.key: k for k in bucket.objects.all()}
    assert keys["new-key"].storage_class == "REDUCED_REDUNDANCY"
    assert keys["the-key"].storage_class == "STANDARD"


@mock_aws
def test_copy_non_existing_file():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    src = "srcbucket"
    target = "target"
    s3_resource.create_bucket(Bucket=src)
    s3_resource.create_bucket(Bucket=target)

    s3_client = boto3.client("s3")
    with pytest.raises(ClientError) as exc:
        s3_client.copy_object(
            Bucket=target, CopySource={"Bucket": src, "Key": "foofoofoo"}, Key="newkey"
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchKey"
    assert err["Message"] == "The specified key does not exist."
    assert err["Key"] == "foofoofoo"


@mock_aws
def test_copy_object_with_versioning():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    client.create_bucket(
        Bucket="blah", CreateBucketConfiguration={"LocationConstraint": "eu-west-1"}
    )
    client.put_bucket_versioning(
        Bucket="blah", VersioningConfiguration={"Status": "Enabled"}
    )

    client.put_object(Bucket="blah", Key="test1", Body=b"test1")
    client.put_object(Bucket="blah", Key="test2", Body=b"test2")

    _ = client.get_object(Bucket="blah", Key="test1")["VersionId"]
    obj2_version = client.get_object(Bucket="blah", Key="test2")["VersionId"]

    client.copy_object(
        CopySource={"Bucket": "blah", "Key": "test1"}, Bucket="blah", Key="test2"
    )
    obj2_version_new = client.get_object(Bucket="blah", Key="test2")["VersionId"]

    # Version should be different to previous version
    assert obj2_version_new != obj2_version

    client.copy_object(
        CopySource={"Bucket": "blah", "Key": "test2", "VersionId": obj2_version},
        Bucket="blah",
        Key="test3",
    )
    obj3_version_new = client.get_object(Bucket="blah", Key="test3")["VersionId"]
    assert obj3_version_new != obj2_version_new

    # Copy file that doesn't exist
    with pytest.raises(ClientError) as exc:
        client.copy_object(
            CopySource={"Bucket": "blah", "Key": "test4", "VersionId": obj2_version},
            Bucket="blah",
            Key="test5",
        )
    assert exc.value.response["Error"]["Code"] == "NoSuchKey"

    response = client.create_multipart_upload(Bucket="blah", Key="test4")
    upload_id = response["UploadId"]
    response = client.upload_part_copy(
        Bucket="blah",
        Key="test4",
        CopySource={"Bucket": "blah", "Key": "test3", "VersionId": obj3_version_new},
        UploadId=upload_id,
        PartNumber=1,
    )
    etag = response["CopyPartResult"]["ETag"]
    client.complete_multipart_upload(
        Bucket="blah",
        Key="test4",
        UploadId=upload_id,
        MultipartUpload={"Parts": [{"ETag": etag, "PartNumber": 1}]},
    )

    response = client.get_object(Bucket="blah", Key="test4")
    data = response["Body"].read()
    assert data == b"test2"


@mock_aws
def test_copy_object_from_unversioned_to_versioned_bucket():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    client.create_bucket(
        Bucket="src", CreateBucketConfiguration={"LocationConstraint": "eu-west-1"}
    )
    client.create_bucket(
        Bucket="dest", CreateBucketConfiguration={"LocationConstraint": "eu-west-1"}
    )
    client.put_bucket_versioning(
        Bucket="dest", VersioningConfiguration={"Status": "Enabled"}
    )

    client.put_object(Bucket="src", Key="test", Body=b"content")

    obj2_version_new = client.copy_object(
        CopySource={"Bucket": "src", "Key": "test"}, Bucket="dest", Key="test"
    ).get("VersionId")

    # VersionId should be present in the response
    assert obj2_version_new is not None


@mock_aws
def test_copy_object_with_replacement_tagging():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    client.create_bucket(Bucket="mybucket")
    client.put_object(
        Bucket="mybucket", Key="original", Body=b"test", Tagging="tag=old"
    )

    # using system tags will fail
    with pytest.raises(ClientError) as err:
        client.copy_object(
            CopySource={"Bucket": "mybucket", "Key": "original"},
            Bucket="mybucket",
            Key="copy1",
            TaggingDirective="REPLACE",
            Tagging="aws:tag=invalid_key",
        )

    exc = err.value
    assert exc.response["Error"]["Code"] == "InvalidTag"

    client.copy_object(
        CopySource={"Bucket": "mybucket", "Key": "original"},
        Bucket="mybucket",
        Key="copy1",
        TaggingDirective="REPLACE",
        Tagging="tag=new",
    )
    client.copy_object(
        CopySource={"Bucket": "mybucket", "Key": "original"},
        Bucket="mybucket",
        Key="copy2",
        TaggingDirective="COPY",
    )

    tags1 = client.get_object_tagging(Bucket="mybucket", Key="copy1")["TagSet"]
    assert tags1 == [{"Key": "tag", "Value": "new"}]
    tags2 = client.get_object_tagging(Bucket="mybucket", Key="copy2")["TagSet"]
    assert tags2 == [{"Key": "tag", "Value": "old"}]


@mock_aws
def test_copy_object_with_kms_encryption():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    kms_client = boto3.client("kms", region_name=DEFAULT_REGION_NAME)
    kms_key = kms_client.create_key()["KeyMetadata"]["KeyId"]

    client.create_bucket(
        Bucket="blah", CreateBucketConfiguration={"LocationConstraint": "eu-west-1"}
    )

    client.put_object(Bucket="blah", Key="test1", Body=b"test1")

    client.copy_object(
        CopySource={"Bucket": "blah", "Key": "test1"},
        Bucket="blah",
        Key="test2",
        SSEKMSKeyId=kms_key,
        ServerSideEncryption="aws:kms",
    )
    result = client.head_object(Bucket="blah", Key="test2")
    assert result["SSEKMSKeyId"] == kms_key
    assert result["ServerSideEncryption"] == "aws:kms"


@mock_aws
def test_copy_object_in_place_with_encryption():
    kms_client = boto3.client("kms", region_name=DEFAULT_REGION_NAME)
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    kms_key = kms_client.create_key()["KeyMetadata"]["KeyId"]
    bucket = s3_resource.Bucket("test_bucket")
    bucket.create()
    key = "source-key"
    resp = client.put_object(
        Bucket="test_bucket",
        Key=key,
        Body=b"somedata",
        ServerSideEncryption="aws:kms",
        BucketKeyEnabled=True,
        SSEKMSKeyId=kms_key,
    )
    assert resp["BucketKeyEnabled"] is True

    # assert that you can copy in place with the same Encryption settings
    client.copy_object(
        Bucket="test_bucket",
        CopySource=f"test_bucket/{key}",
        Key=key,
        ServerSideEncryption="aws:kms",
        BucketKeyEnabled=True,
        SSEKMSKeyId=kms_key,
    )

    # assert that the BucketKeyEnabled setting is not kept in the destination key
    resp = client.copy_object(
        Bucket="test_bucket",
        CopySource=f"test_bucket/{key}",
        Key=key,
        ServerSideEncryption="aws:kms",
        SSEKMSKeyId=kms_key,
    )
    assert "BucketKeyEnabled" not in resp

    # This is an edge case, if the source object SSE was not AES256,
    # AWS allows you to not specify any fields as it will use AES256 by
    # default and is different from the source key.
    resp = client.copy_object(
        Bucket="test_bucket",
        CopySource=f"test_bucket/{key}",
        Key=key,
    )
    assert resp["ServerSideEncryption"] == "AES256"

    # Check that it allows copying in the place with the same
    # ServerSideEncryption setting as the source.
    resp = client.copy_object(
        Bucket="test_bucket",
        CopySource=f"test_bucket/{key}",
        Key=key,
        ServerSideEncryption="AES256",
    )
    assert resp["ServerSideEncryption"] == "AES256"


@mock_aws
def test_copy_object_in_place_with_storage_class():
    """Validate setting StorageClass allows a copy in place.

    This should be true even if destination object is the same as source.
    """
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "test-bucket"
    bucket = s3_resource.Bucket(bucket_name)
    bucket.create()
    key = "source-key"
    bucket.put_object(Key=key, Body=b"somedata", StorageClass="STANDARD")
    client.copy_object(
        Bucket=bucket_name,
        CopySource=f"{bucket_name}/{key}",
        Key=key,
        StorageClass="STANDARD",
    )
    # verify that the copy worked
    resp = client.get_object_attributes(
        Bucket=bucket_name, Key=key, ObjectAttributes=["StorageClass"]
    )
    assert resp["StorageClass"] == "STANDARD"


@mock_aws
def test_copy_object_does_not_copy_storage_class():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket = s3_resource.Bucket("test_bucket")
    bucket.create()
    source_key = "source-key"
    dest_key = "dest-key"
    bucket.put_object(Key=source_key, Body=b"somedata", StorageClass="STANDARD_IA")
    client.copy_object(
        Bucket="test_bucket",
        CopySource=f"test_bucket/{source_key}",
        Key=dest_key,
    )

    # Verify that the destination key does not have STANDARD_IA as StorageClass
    keys = {k.key: k for k in bucket.objects.all()}
    assert keys[source_key].storage_class == "STANDARD_IA"
    assert keys[dest_key].storage_class == "STANDARD"


@mock_aws
def test_copy_object_does_not_copy_acl():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "testbucket"
    bucket = s3_resource.Bucket(bucket_name)
    bucket.create()
    source_key = "source-key"
    dest_key = "dest-key"
    control_key = "control-key"
    # do not set ACL for the control key to get default ACL
    bucket.put_object(Key=control_key, Body=b"somedata")
    # set ACL for the source key to check if it will get copied
    bucket.put_object(Key=source_key, Body=b"somedata", ACL="public-read")
    # copy object without specifying ACL, so it should get default ACL
    client.copy_object(
        Bucket=bucket_name,
        CopySource=f"{bucket_name}/{source_key}",
        Key=dest_key,
    )

    # Get the ACL from the all the keys
    source_acl = client.get_object_acl(Bucket=bucket_name, Key=source_key)
    dest_acl = client.get_object_acl(Bucket=bucket_name, Key=dest_key)
    default_acl = client.get_object_acl(Bucket=bucket_name, Key=control_key)
    # assert that the source key ACL are different from the destination key ACL
    assert source_acl["Grants"] != dest_acl["Grants"]
    # assert that the copied key got the default ACL like the control key
    assert default_acl["Grants"] == dest_acl["Grants"]


@mock_aws
def test_copy_object_in_place_with_metadata():
    s3_resource = boto3.resource("s3", region_name=DEFAULT_REGION_NAME)
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "testbucket"
    bucket = s3_resource.Bucket(bucket_name)
    bucket.create()
    key_name = "source-key"
    bucket.put_object(Key=key_name, Body=b"somedata")

    # test that giving metadata is not enough and should provide
    # MetadataDirective=REPLACE on top.
    with pytest.raises(ClientError) as exc:
        client.copy_object(
            Bucket=bucket_name,
            CopySource=f"{bucket_name}/{key_name}",
            Key=key_name,
            Metadata={"key": "value"},
        )
        assert exc.value.response["Error"]["Message"] == (
            "This copy request is illegal because it is trying to copy an "
            "object to itself without changing the object's metadata, "
            "storage class, website redirect location or encryption attributes."
        )

    # you can only provide MetadataDirective=REPLACE and it will copy without any metadata
    client.copy_object(
        Bucket=bucket_name,
        CopySource=f"{bucket_name}/{key_name}",
        Key=key_name,
        MetadataDirective="REPLACE",
    )

    result = client.head_object(Bucket=bucket_name, Key=key_name)
    assert result["Metadata"] == {}


@mock_aws
def test_copy_objet_legal_hold():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "testbucket"
    source_key = "source-key"
    dest_key = "dest-key"
    client.create_bucket(Bucket=bucket_name, ObjectLockEnabledForBucket=True)
    client.put_object(
        Bucket=bucket_name,
        Key=source_key,
        Body=b"somedata",
        ObjectLockLegalHoldStatus="ON",
    )

    head_object = client.head_object(Bucket=bucket_name, Key=source_key)
    assert head_object["ObjectLockLegalHoldStatus"] == "ON"
    assert "VersionId" in head_object
    version_id = head_object["VersionId"]

    resp = client.copy_object(
        Bucket=bucket_name,
        CopySource=f"{bucket_name}/{source_key}",
        Key=dest_key,
    )
    assert resp["CopySourceVersionId"] == version_id
    assert resp["VersionId"] != version_id

    # the destination key did not keep the legal hold from the source key
    head_object = client.head_object(Bucket=bucket_name, Key=dest_key)
    assert "ObjectLockLegalHoldStatus" not in head_object


@mock_aws
def test_s3_copy_object_lock():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "testbucket"
    source_key = "source-key"
    dest_key = "dest-key"
    client.create_bucket(Bucket=bucket_name, ObjectLockEnabledForBucket=True)
    # manipulate a bit the datetime object for an easier comparison
    retain_until = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(
        minutes=1
    )
    retain_until = retain_until.replace(microsecond=0)

    client.put_object(
        Bucket=bucket_name,
        Key=source_key,
        Body="test",
        ObjectLockMode="GOVERNANCE",
        ObjectLockRetainUntilDate=retain_until,
    )

    head_object = client.head_object(Bucket=bucket_name, Key=source_key)

    assert head_object["ObjectLockMode"] == "GOVERNANCE"
    assert head_object["ObjectLockRetainUntilDate"] == retain_until
    assert "VersionId" in head_object
    version_id = head_object["VersionId"]

    resp = client.copy_object(
        Bucket=bucket_name,
        CopySource=f"{bucket_name}/{source_key}",
        Key=dest_key,
    )
    assert resp["CopySourceVersionId"] == version_id
    assert resp["VersionId"] != version_id

    # the destination key did not keep the lock mode nor the lock until from the source key
    head_object = client.head_object(Bucket=bucket_name, Key=dest_key)
    assert "ObjectLockMode" not in head_object
    assert "ObjectLockRetainUntilDate" not in head_object


@mock_aws
def test_copy_object_in_place_website_redirect_location():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "testbucket"
    key = "source-key"
    client.create_bucket(Bucket=bucket_name)
    # This test will validate that setting WebsiteRedirectLocation
    # (even the same as source) allows a copy in place.

    client.put_object(
        Bucket=bucket_name,
        Key=key,
        Body="test",
        WebsiteRedirectLocation="/test/direct",
    )

    head_object = client.head_object(Bucket=bucket_name, Key=key)
    assert head_object["WebsiteRedirectLocation"] == "/test/direct"

    # copy the object with the same WebsiteRedirectLocation as the source object
    client.copy_object(
        Bucket=bucket_name,
        CopySource=f"{bucket_name}/{key}",
        Key=key,
        WebsiteRedirectLocation="/test/direct",
    )

    head_object = client.head_object(Bucket=bucket_name, Key=key)
    assert head_object["WebsiteRedirectLocation"] == "/test/direct"


@mock_aws
def test_copy_object_in_place_with_bucket_encryption():
    # If a bucket has encryption configured, it will allow copy in place per default
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "test-bucket"
    client.create_bucket(Bucket=bucket_name)
    key = "source-key"

    response = client.put_bucket_encryption(
        Bucket=bucket_name,
        ServerSideEncryptionConfiguration={
            "Rules": [
                {
                    "ApplyServerSideEncryptionByDefault": {"SSEAlgorithm": "AES256"},
                    "BucketKeyEnabled": False,
                },
            ]
        },
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = client.put_object(
        Body=b"",
        Bucket=bucket_name,
        Key=key,
    )
    assert response["ServerSideEncryption"] == "AES256"

    response = client.copy_object(
        Bucket=bucket_name,
        CopySource={"Bucket": bucket_name, "Key": key},
        Key=key,
    )
    assert response["ServerSideEncryption"] == "AES256"


@mock_aws
def test_copy_object_in_place_with_versioning():
    # If a bucket has versioning enabled, it will allow copy in place
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "testbucket"
    client.create_bucket(Bucket=bucket_name)
    key = "source-key"

    response = client.put_object(
        Body=b"",
        Bucket=bucket_name,
        Key=key,
    )

    response = client.put_bucket_versioning(
        Bucket=bucket_name,
        VersioningConfiguration={
            "MFADelete": "Disabled",
            "Status": "Enabled",
        },
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = client.put_object(
        Body=b"",
        Bucket=bucket_name,
        Key=key,
    )
    version_id = response["ResponseMetadata"]["HTTPHeaders"]["x-amz-version-id"]
    assert version_id and version_id != "null"

    response = client.copy_object(
        Bucket=bucket_name,
        CopySource={"Bucket": bucket_name, "Key": key, "VersionId": version_id},
        Key=key,
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = client.copy_object(
        Bucket=bucket_name,
        CopySource={"Bucket": bucket_name, "Key": key, "VersionId": "null"},
        Key=key,
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = client.list_object_versions(
        Bucket=bucket_name,
        Prefix=key,
    )
    assert len(response["Versions"]) == 4


@mock_aws
@pytest.mark.parametrize(
    "algorithm",
    ["CRC32", "SHA1", "SHA256"],
)
def test_copy_key_boto3_with_both_sha256_checksum(algorithm):
    """Validate that moto S3 checksum calculations are correct.

    We first create an object with a Checksum calculated by boto, by
    specifying ChecksumAlgorithm="SHA256".

    We then retrieve the right checksum from this request.

    We copy the object while requesting moto to recalculate the checksum
    for that key.

    We verify that both checksums are equal.
    """
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    source_key = "source-key"
    dest_key = "dest-key"
    bucket = "foobar"
    body = b"checksum-test"
    client.create_bucket(Bucket=bucket)

    checksum_key = f"Checksum{algorithm}"

    resp = client.put_object(
        Bucket=bucket,
        Key=source_key,
        Body=body,
        ChecksumAlgorithm=algorithm,
    )
    assert checksum_key in resp
    checksum_by_boto = resp[checksum_key]

    resp = client.copy_object(
        Bucket=bucket,
        CopySource=f"{bucket}/{source_key}",
        Key=dest_key,
        ChecksumAlgorithm=algorithm,
    )

    assert checksum_key in resp["CopyObjectResult"]
    assert resp["CopyObjectResult"][checksum_key] == checksum_by_boto


@mock_aws
@pytest.mark.parametrize(
    "algorithm, checksum",
    [
        ("CRC32", "lVk/nw=="),
        ("SHA1", "jbXkHAsXUrubtL3dqDQ4w+7WXc0="),
        ("SHA256", "1YQo81vx2VFUl0q5ccWISq8AkSBQQ0WO80S82TmfdIQ="),
    ],
)
def test_copy_object_calculates_checksum(algorithm, checksum):
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    source_key = "source-key"
    dest_key = "dest-key"
    bucket = "foobar"
    body = b"test-checksum"
    client.create_bucket(Bucket=bucket)

    checksum_key = f"Checksum{algorithm}"

    resp = client.put_object(
        Bucket=bucket,
        Key=source_key,
        Body=body,
    )
    assert checksum_key not in resp

    resp = client.copy_object(
        Bucket=bucket,
        CopySource=f"{bucket}/{source_key}",
        Key=dest_key,
        ChecksumAlgorithm=algorithm,
    )

    assert checksum_key in resp["CopyObjectResult"]
    assert resp["CopyObjectResult"][checksum_key] == checksum


@mock_aws
def test_copy_object_keeps_checksum():
    client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    source_key = "source-key"
    dest_key = "dest-key"
    bucket = "foobar"
    body = b"test-checksum"
    expected_checksum = "1YQo81vx2VFUl0q5ccWISq8AkSBQQ0WO80S82TmfdIQ="
    client.create_bucket(Bucket=bucket)

    # put an object with a checksum
    resp = client.put_object(
        Bucket=bucket,
        Key=source_key,
        Body=body,
        ChecksumAlgorithm="SHA256",
    )
    assert "ChecksumSHA256" in resp
    assert resp["ChecksumSHA256"] == expected_checksum

    # do not specify the checksum
    resp = client.copy_object(
        Bucket=bucket,
        CopySource=f"{bucket}/{source_key}",
        Key=dest_key,
    )

    # assert that it kept the checksum from the source key
    assert "ChecksumSHA256" in resp["CopyObjectResult"]
    assert resp["CopyObjectResult"]["ChecksumSHA256"] == expected_checksum



import json

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# https://docs.aws.amazon.com/systems-manager/latest/userguide/integration-ps-secretsmanager.html


@mock_aws
def test_get_value_from_secrets_manager__by_name():
    # given
    ssm = boto3.client("ssm", "eu-north-1")
    secrets_manager = boto3.client("secretsmanager", "eu-north-1")
    secret_name = "mysecret"
    # when
    secrets_manager.create_secret(Name=secret_name, SecretString="some secret")
    # then
    param = ssm.get_parameter(
        Name=f"/aws/reference/secretsmanager/{secret_name}", WithDecryption=True
    )["Parameter"]
    assert param["Name"] == "mysecret"
    assert param["Type"] == "SecureString"
    assert param["Value"] == "some secret"
    assert param["Version"] == 0
    assert "SourceResult" in param

    secret = secrets_manager.describe_secret(SecretId=secret_name)
    source_result = json.loads(param["SourceResult"])

    assert source_result["ARN"] == secret["ARN"]
    assert source_result["Name"] == secret["Name"]
    assert source_result["VersionIdsToStages"] == secret["VersionIdsToStages"]


@mock_aws
def test_get_value_from_secrets_manager__without_decryption():
    # Note that the parameter does not need to exist
    ssm = boto3.client("ssm", "eu-north-1")
    with pytest.raises(ClientError) as exc:
        ssm.get_parameter(Name="/aws/reference/secretsmanager/sth")
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert err["Message"] == (
        "WithDecryption flag must be True for retrieving a Secret Manager secret."
    )


@mock_aws
def test_get_value_from_secrets_manager__with_decryption_false():
    # Note that the parameter does not need to exist
    ssm = boto3.client("ssm", "eu-north-1")
    with pytest.raises(ClientError) as exc:
        ssm.get_parameter(
            Name="/aws/reference/secretsmanager/sth", WithDecryption=False
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert err["Message"] == (
        "WithDecryption flag must be True for retrieving a Secret Manager secret."
    )


@mock_aws
def test_get_value_from_secrets_manager__by_id():
    # given
    ssm = boto3.client("ssm", "eu-north-1")
    secrets_manager = boto3.client("secretsmanager", "eu-north-1")
    name = "mysecret"
    # when
    r1 = secrets_manager.create_secret(Name=name, SecretString="1st")
    version_id1 = r1["VersionId"]
    secrets_manager.put_secret_value(
        SecretId=name, SecretString="2nd", VersionStages=["AWSCURRENT"]
    )
    r3 = secrets_manager.put_secret_value(
        SecretId=name, SecretString="3rd", VersionStages=["ST1"]
    )
    version_id3 = r3["VersionId"]
    # then
    full_name = f"/aws/reference/secretsmanager/{name}:{version_id1}"
    param = ssm.get_parameter(Name=full_name, WithDecryption=True)["Parameter"]
    assert param["Value"] == "1st"

    full_name = f"/aws/reference/secretsmanager/{name}"
    param = ssm.get_parameter(Name=full_name, WithDecryption=True)["Parameter"]
    assert param["Value"] == "2nd"

    full_name = f"/aws/reference/secretsmanager/{name}:{version_id3}"
    param = ssm.get_parameter(Name=full_name, WithDecryption=True)["Parameter"]
    assert param["Value"] == "3rd"


@mock_aws
def test_get_value_from_secrets_manager__by_version():
    # given
    ssm = boto3.client("ssm", "eu-north-1")
    secrets_manager = boto3.client("secretsmanager", "eu-north-1")
    name = "mysecret"
    # when
    secrets_manager.create_secret(Name=name, SecretString="1st")
    secrets_manager.put_secret_value(
        SecretId=name, SecretString="2nd", VersionStages=["AWSCURRENT"]
    )
    # then
    full_name = f"/aws/reference/secretsmanager/{name}:AWSPREVIOUS"
    param = ssm.get_parameter(Name=full_name, WithDecryption=True)["Parameter"]
    assert param["Value"] == "1st"


@mock_aws
def test_get_value_from_secrets_manager__param_does_not_exist():
    ssm = boto3.client("ssm", "us-east-1")
    with pytest.raises(ClientError) as exc:
        ssm.get_parameter(
            Name="/aws/reference/secretsmanager/test", WithDecryption=True
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ParameterNotFound"
    assert err["Message"] == (
        "An error occurred (ParameterNotFound) when referencing Secrets "
        "Manager: Secret /aws/reference/secretsmanager/test not found."
    )



from __future__ import annotations

import json
import logging
from typing import Final, List, Optional

from moto.stepfunctions.parser.asl.component.common.parameters import Parameters
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_selector import (
    ItemSelector,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.inline_iteration_component import (
    InlineIterationComponent,
    InlineIterationComponentEvalInput,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.iterator_decl import (
    IteratorDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.iterator_worker import (
    IteratorWorker,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment

LOG = logging.getLogger(__name__)


class IteratorEvalInput(InlineIterationComponentEvalInput):
    parameters: Final[Optional[ItemSelector]]

    def __init__(
        self,
        state_name: str,
        max_concurrency: int,
        input_items: List[json],
        parameters: Optional[Parameters],
    ):
        super().__init__(
            state_name=state_name,
            max_concurrency=max_concurrency,
            input_items=input_items,
        )
        self.parameters = parameters


class Iterator(InlineIterationComponent):
    _eval_input: Optional[IteratorEvalInput]

    def _create_worker(self, env: Environment) -> IteratorWorker:
        return IteratorWorker(
            work_name=self._eval_input.state_name,
            job_pool=self._job_pool,
            env=env,
            parameters=self._eval_input.parameters,
        )

    @classmethod
    def from_declaration(cls, iterator_decl: IteratorDecl):
        return cls(
            start_at=iterator_decl.start_at,
            states=iterator_decl.states,
            comment=iterator_decl.comment,
        )



from typing import Final

from moto.stepfunctions.parser.asl.component.intrinsic.member import Member


class MemberAccess(Member):
    def __init__(self, subject: Member, target: Member):
        self.subject: Final[Member] = subject
        self.target: Final[Member] = target



from typing import Any

from moto.core.exceptions import AWSError, JsonRESTError, RESTError

ERROR_WITH_MODEL_NAME = """{% extends 'single_error' %}
{% block extra %}<ModelName>{{ model }}</ModelName>{% endblock %}
"""


class SagemakerClientError(RESTError):
    extended_templates = {"model_error": ERROR_WITH_MODEL_NAME}
    env = RESTError.extended_environment(extended_templates)

    def __init__(self, *args: Any, **kwargs: Any):
        kwargs.setdefault("template", "single_error")
        super().__init__(*args, **kwargs)


class ModelError(RESTError):
    extended_templates = {"model_error": ERROR_WITH_MODEL_NAME}
    env = RESTError.extended_environment(extended_templates)

    def __init__(self, *args: Any, **kwargs: Any):
        kwargs.setdefault("template", "model_error")
        super().__init__(*args, **kwargs)


class MissingModel(ModelError):
    code = 404

    def __init__(self, model: str):
        super().__init__("NoSuchModel", "Could not find model", model=model)


class ValidationError(JsonRESTError):
    def __init__(self, message: str):
        super().__init__("ValidationException", message)


class AWSValidationException(AWSError):
    TYPE = "ValidationException"


class ResourceInUseException(AWSError):
    TYPE = "ResourceInUse"


class ResourceNotFound(JsonRESTError):
    def __init__(self, message: str):
        super().__init__(__class__.__name__, message)  # type: ignore



import pytest

from moto.logs.logs_query.query_parser import parse_query


@pytest.mark.parametrize(
    "query,fields,limit,sort",
    [
        ("fields @timestamp", ["@timestamp"], None, []),
        ("fields @timestamp, @message", ["@timestamp", "@message"], None, []),
        ("limit 42", [], 42, []),
        ("sort @timestamp desc", [], None, [("@timestamp", "desc")]),
        ("sort @timestamp asc", [], None, [("@timestamp", "asc")]),
        ("sort @timestamp", [], None, [("@timestamp", "desc")]),
        ("fields @timestamp | limit 42", ["@timestamp"], 42, []),
        ("limit 42 | fields @timestamp", ["@timestamp"], 42, []),
        ("fields @fld | sort @fld | limit 42", ["@fld"], 42, [("@fld", "desc")]),
        ("sort @fld asc | fields @fld | limit 42", ["@fld"], 42, [("@fld", "asc")]),
        ("limit 42 | sort @fld | fields @fld", ["@fld"], 42, [("@fld", "desc")]),
    ],
)
def test_query(query, fields, limit, sort):
    parsed = parse_query(query)
    assert parsed.fields == fields
    assert parsed.limit == limit
    assert parsed.sort == sort



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

DEFAULT_REGION = "us-west-2"


@mock_aws
def test_create_db_proxy():
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    ec2_client = boto3.client("ec2", region_name=DEFAULT_REGION)
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(CidrBlock="10.0.1.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    subnet_id_2 = ec2_client.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    resp = rds_client.create_db_proxy(
        DBProxyName="testrdsproxy",
        EngineFamily="MYSQL",
        Auth=[
            {
                "Description": "Test Description",
                "UserName": "Test Username",
                "AuthScheme": "SECRETS",
                "SecretArn": "TestSecretARN",
                "IAMAuth": "ENABLED",
                "ClientPasswordAuthType": "MYSQL_NATIVE_PASSWORD",
            },
        ],
        RoleArn="TestArn",
        VpcSubnetIds=[subnet_id, subnet_id_2],
        RequireTLS=True,
        Tags=[{"Key": "TestKey", "Value": "TestValue"}],
    )
    db_proxy = resp["DBProxy"]
    assert db_proxy["DBProxyName"] == "testrdsproxy"
    assert (
        db_proxy["DBProxyArn"]
        == f"arn:aws:rds:us-west-2:{ACCOUNT_ID}:db-proxy:testrdsproxy"
    )
    assert db_proxy["Status"] == "availible"
    assert db_proxy["EngineFamily"] == "MYSQL"
    assert db_proxy["VpcId"] == vpc_id
    assert db_proxy["VpcSecurityGroupIds"] == []
    assert db_proxy["VpcSubnetIds"] == [subnet_id, subnet_id_2]
    assert db_proxy["Auth"] == [
        {
            "Description": "Test Description",
            "UserName": "Test Username",
            "AuthScheme": "SECRETS",
            "SecretArn": "TestSecretARN",
            "IAMAuth": "ENABLED",
            "ClientPasswordAuthType": "MYSQL_NATIVE_PASSWORD",
        }
    ]
    assert db_proxy["RoleArn"] == "TestArn"
    assert db_proxy["RequireTLS"] is True
    assert db_proxy["IdleClientTimeout"] == 1800
    assert db_proxy["DebugLogging"] is False


@mock_aws
def test_describe_db_proxies():
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    ec2_client = boto3.client("ec2", region_name=DEFAULT_REGION)
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(CidrBlock="10.0.1.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    subnet_id_2 = ec2_client.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    rds_client.create_db_proxy(
        DBProxyName="testrdsproxydescribe",
        EngineFamily="MYSQL",
        Auth=[
            {
                "Description": "Test Description",
                "UserName": "Test Username",
                "AuthScheme": "SECRETS",
                "SecretArn": "TestSecretARN",
                "IAMAuth": "ENABLED",
                "ClientPasswordAuthType": "MYSQL_NATIVE_PASSWORD",
            },
        ],
        RoleArn="TestArn",
        VpcSubnetIds=[subnet_id, subnet_id_2],
        RequireTLS=True,
        Tags=[
            {"Key": "TestKey", "Value": "TestValue"},
            {"Key": "aaa", "Value": "bbb"},
        ],
    )
    response = rds_client.describe_db_proxies(DBProxyName="testrdsproxydescribe")
    db_proxy = response["DBProxies"][0]
    assert db_proxy["DBProxyName"] == "testrdsproxydescribe"
    assert (
        db_proxy["DBProxyArn"]
        == f"arn:aws:rds:us-west-2:{ACCOUNT_ID}:db-proxy:testrdsproxydescribe"
    )
    assert db_proxy["Status"] == "availible"
    assert db_proxy["EngineFamily"] == "MYSQL"
    assert db_proxy["VpcId"] == vpc_id
    assert db_proxy["VpcSecurityGroupIds"] == []
    assert db_proxy["VpcSubnetIds"] == [subnet_id, subnet_id_2]
    assert db_proxy["Auth"] == [
        {
            "Description": "Test Description",
            "UserName": "Test Username",
            "AuthScheme": "SECRETS",
            "SecretArn": "TestSecretARN",
            "IAMAuth": "ENABLED",
            "ClientPasswordAuthType": "MYSQL_NATIVE_PASSWORD",
        }
    ]
    assert db_proxy["RoleArn"] == "TestArn"
    assert db_proxy["RequireTLS"] is True
    assert db_proxy["IdleClientTimeout"] == 1800
    assert db_proxy["DebugLogging"] is False


@mock_aws
def test_list_tags_db_proxy():
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    ec2_client = boto3.client("ec2", region_name=DEFAULT_REGION)
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(CidrBlock="10.0.1.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    subnet_id_2 = ec2_client.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    resp = rds_client.create_db_proxy(
        DBProxyName="testrdsproxydescribe",
        EngineFamily="MYSQL",
        Auth=[
            {
                "Description": "Test Description",
                "UserName": "Test Username",
                "AuthScheme": "SECRETS",
                "SecretArn": "TestSecretARN",
                "IAMAuth": "ENABLED",
                "ClientPasswordAuthType": "MYSQL_NATIVE_PASSWORD",
            },
        ],
        RoleArn="TestArn",
        VpcSubnetIds=[subnet_id, subnet_id_2],
        RequireTLS=True,
        Tags=[
            {"Key": "TestKey", "Value": "TestValue"},
            {"Key": "aaa", "Value": "bbb"},
        ],
    )
    arn = resp["DBProxy"]["DBProxyArn"]
    resp = rds_client.list_tags_for_resource(ResourceName=arn)
    assert resp["TagList"] == [
        {"Value": "TestValue", "Key": "TestKey"},
        {"Value": "bbb", "Key": "aaa"},
    ]


@mock_aws
def test_create_db_proxy_invalid_subnet():
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    ec2_client = boto3.client("ec2", region_name=DEFAULT_REGION)
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    vpc_id_2 = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(CidrBlock="10.0.1.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    subnet_id_2 = ec2_client.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc_id_2)[
        "Subnet"
    ]["SubnetId"]
    with pytest.raises(ClientError) as ex:
        rds_client.create_db_proxy(
            DBProxyName="testrdsproxy",
            EngineFamily="MYSQL",
            Auth=[
                {
                    "Description": "Test Description",
                    "UserName": "Test Username",
                    "AuthScheme": "SECRETS",
                    "SecretArn": "TestSecretARN",
                    "IAMAuth": "ENABLED",
                    "ClientPasswordAuthType": "MYSQL_NATIVE_PASSWORD",
                },
            ],
            RoleArn="TestArn",
            VpcSubnetIds=[subnet_id, subnet_id_2],
            RequireTLS=True,
            Tags=[{"Key": "TestKey", "Value": "TestValue"}],
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "InvalidSubnet"


@mock_aws
def test_create_db_proxy_duplicate_name():
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    ec2_client = boto3.client("ec2", region_name=DEFAULT_REGION)
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(CidrBlock="10.0.1.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    subnet_id_2 = ec2_client.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc_id)[
        "Subnet"
    ]["SubnetId"]
    rds_client.create_db_proxy(
        DBProxyName="testrdsproxy",
        EngineFamily="MYSQL",
        Auth=[
            {
                "Description": "Test Description",
                "UserName": "Test Username",
                "AuthScheme": "SECRETS",
                "SecretArn": "TestSecretARN",
                "IAMAuth": "ENABLED",
                "ClientPasswordAuthType": "MYSQL_NATIVE_PASSWORD",
            },
        ],
        RoleArn="TestArn",
        VpcSubnetIds=[subnet_id, subnet_id_2],
        RequireTLS=True,
        Tags=[{"Key": "TestKey", "Value": "TestValue"}],
    )
    with pytest.raises(ClientError) as ex:
        rds_client.create_db_proxy(
            DBProxyName="testrdsproxy",
            EngineFamily="MYSQL",
            Auth=[
                {
                    "Description": "Test Description",
                    "UserName": "Test Username",
                    "AuthScheme": "SECRETS",
                    "SecretArn": "TestSecretARN",
                    "IAMAuth": "ENABLED",
                    "ClientPasswordAuthType": "MYSQL_NATIVE_PASSWORD",
                },
            ],
            RoleArn="TestArn",
            VpcSubnetIds=[subnet_id, subnet_id_2],
            RequireTLS=True,
            Tags=[{"Key": "TestKey", "Value": "TestValue"}],
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "DBProxyAlreadyExistsFault"
    assert (
        err["Message"]
        == "Cannot create the DBProxy because a DBProxy with the identifier testrdsproxy already exists."
    )


@mock_aws
def test_describe_db_proxies_not_found():
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    with pytest.raises(ClientError) as ex:
        rds_client.describe_db_proxies(DBProxyName="testrdsproxydescribe")
    err = ex.value.response["Error"]
    assert err["Code"] == "DBProxyNotFoundFault"
    assert (
        err["Message"]
        == "The specified proxy name testrdsproxydescribe doesn't correspond to a proxy owned by your Amazon Web Services account in the specified Amazon Web Services Region."
    )



"""Unit tests for bedrockagent-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

DEFAULT_REGION = "us-east-1"

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_agent():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)

    resp = client.create_agent(
        agentName="agent_name",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )
    # resp = client.create_agent(
    #     agentName="agent_name",
    #     clientToken="client_tokenclient_tokenclient_token",
    #     instruction="instructioninstructioninstructioninstruction",
    #     foundationModel="foundation_model",
    #     description="description",
    #     idleSessionTTLInSeconds=60,
    #     agentResourceRoleArn="agent_resource_role_arn",
    #     customerEncryptionKeyArn="customer_encryption_key_arn",
    #     tags={
    #             "Key": "test-tag-key",
    #         },
    #     promptOverrideConfiguration={
    #     'promptConfigurations': [
    #         {
    #             'promptType': 'PRE_PROCESSING',
    #             'promptCreationMode': 'DEFAULT',
    #             'promptState': 'ENABLED',
    #             'basePromptTemplate': 'string',
    #             'inferenceConfiguration': {
    #                 'temperature': 1.0,
    #                 'topP': 1.0,
    #                 'topK': 123,
    #                 'maximumLength': 123,
    #                 'stopSequences': [
    #                     'string',
    #                 ]
    #             },
    #             'parserMode': 'DEFAULT'
    #         },
    #     ],
    #     'overrideLambda': 'overrideLambdaoverrideLambdaoverrideLambdaoverrideLambda'
    # }
    # )
    assert resp["agent"]["agentName"] == "agent_name"


@mock_aws
def test_get_agent():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    agent = client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )
    resp = client.get_agent(agentId=agent["agent"]["agentId"])
    assert resp["agent"]["agentName"] == "testname"


@mock_aws
def test_get_agent_not_found():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )
    with pytest.raises(ClientError) as e:
        client.get_agent(agentId="non-existent-agent-id")
    assert e.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_agents():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    client.create_agent(
        agentName="testname1",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )
    client.create_agent(
        agentName="testname2",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )
    resp = client.list_agents()
    # assert resp['agent']['agentName'] == "testname"

    assert len(resp["agentSummaries"]) == 2
    assert resp["agentSummaries"][0]["agentName"] == "testname1"
    assert resp["agentSummaries"][1]["agentName"] == "testname2"


@mock_aws
def test_delete_agent():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    agent = client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
        tags={"Key": "test-tag-key"},
    )
    agent_id = agent["agent"]["agentId"]
    resp = client.delete_agent(agentId=agent_id, skipResourceInUseCheck=True)

    assert resp["agentId"] == agent_id
    assert resp["agentStatus"] == "DELETING"


@mock_aws
def test_create_knowledge_base():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    resp = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )

    assert resp["knowledgeBase"]["name"] == "testkb"


@mock_aws
def test_list_knowledge_bases():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    client.create_knowledge_base(
        name="testkb1",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )
    client.create_knowledge_base(
        name="testkb2",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )
    resp = client.list_knowledge_bases()

    assert len(resp["knowledgeBaseSummaries"]) == 2
    assert resp["knowledgeBaseSummaries"][0]["name"] == "testkb1"
    assert resp["knowledgeBaseSummaries"][1]["name"] == "testkb2"


@mock_aws
def test_delete_knowledge_base():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    kb = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )
    kb_id = kb["knowledgeBase"]["knowledgeBaseId"]
    resp = client.delete_knowledge_base(knowledgeBaseId=kb_id)
    assert resp["knowledgeBaseId"] == kb_id
    assert resp["status"] == "DELETING"


@mock_aws
def test_delete_knowledge_base_not_found():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )
    with pytest.raises(ClientError) as e:
        client.delete_knowledge_base(knowledgeBaseId="non-existent-kb-id")
    assert e.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_get_knowledge_base():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    kb = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )
    kb_id = kb["knowledgeBase"]["knowledgeBaseId"]
    resp = client.get_knowledge_base(knowledgeBaseId=kb_id)
    assert resp["knowledgeBase"]["name"] == "testkb"


@mock_aws
def test_get_knowledge_base_not_found():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )
    with pytest.raises(ClientError) as e:
        client.get_knowledge_base(knowledgeBaseId="non-existent-kb-id")
    assert e.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_tag_resource_agent():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    agent = client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
    )
    agent_arn = agent["agent"]["agentArn"]
    resp = client.tag_resource(resourceArn=agent_arn, tags={"Key": "test-tag"})
    resp = client.list_tags_for_resource(resourceArn=agent_arn)
    assert resp["tags"]["Key"] == "test-tag"


@mock_aws
def test_tag_resource_knowledge_base():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    kb = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
    )
    kb_arn = kb["knowledgeBase"]["knowledgeBaseArn"]
    resp = client.tag_resource(resourceArn=kb_arn, tags={"Key": "test-tag"})
    resp = client.list_tags_for_resource(resourceArn=kb_arn)
    assert resp["tags"]["Key"] == "test-tag"


@mock_aws
def test_untag_resource_agent():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    agent = client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
    )
    agent_arn = agent["agent"]["agentArn"]
    resp = client.tag_resource(
        resourceArn=agent_arn, tags={"Key1": "test-tag", "Key2": "test-tag2"}
    )
    resp = client.untag_resource(resourceArn=agent_arn, tagKeys=["Key1"])
    resp = client.list_tags_for_resource(resourceArn=agent_arn)
    assert len(resp["tags"]) == 1
    assert resp["tags"]["Key2"] == "test-tag2"


@mock_aws
def test_untag_resource_knowledge_base():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    kb = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
    )
    kb_arn = kb["knowledgeBase"]["knowledgeBaseArn"]
    resp = client.tag_resource(
        resourceArn=kb_arn, tags={"Key1": "test-tag", "Key2": "test-tag2"}
    )
    resp = client.untag_resource(resourceArn=kb_arn, tagKeys=["Key1", "Key2"])
    resp = client.list_tags_for_resource(resourceArn=kb_arn)
    assert len(resp["tags"]) == 0


@mock_aws
def test_list_tags_for_resource_agent():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    agent = client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
    )
    agent_arn = agent["agent"]["agentArn"]
    resp = client.tag_resource(
        resourceArn=agent_arn, tags={"Key1": "test-tag", "Key2": "test-tag2"}
    )
    resp = client.list_tags_for_resource(resourceArn=agent_arn)
    assert resp["tags"]["Key1"] == "test-tag"
    assert resp["tags"]["Key2"] == "test-tag2"


@mock_aws
def test_list_tags_for_resource_knowledge_base():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    kb = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
    )
    kb_arn = kb["knowledgeBase"]["knowledgeBaseArn"]
    resp = client.tag_resource(
        resourceArn=kb_arn, tags={"Key1": "test-tag", "Key2": "test-tag2"}
    )
    resp = client.list_tags_for_resource(resourceArn=kb_arn)
    assert resp["tags"]["Key1"] == "test-tag"
    assert resp["tags"]["Key2"] == "test-tag2"


@mock_aws
def test_create_knowledge_base_bad_knowledge_base_config():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    with pytest.raises(ClientError) as e:
        client.create_knowledge_base(
            name="testkb",
            description="description",
            roleArn="test_role_arn",
            knowledgeBaseConfiguration={
                "type": "vECTOR",
            },
            storageConfiguration={
                "type": "OPENSEARCH_SERVERLESS",
            },
            tags={"Key": "test-tag"},
        )
    assert e.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_create_knowledge_base_bad_storage_config():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    with pytest.raises(ClientError) as e:
        client.create_knowledge_base(
            name="testkb",
            description="description",
            roleArn="test_role_arn",
            knowledgeBaseConfiguration={
                "type": "VECTOR",
            },
            storageConfiguration={
                "type": "oPENSEARCH_SERVERLES",
            },
            tags={"Key": "test-tag"},
        )
    assert e.value.response["Error"]["Code"] == "ValidationException"


# @mock_aws
# def test_list_agents_token():
#     client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
#     client.create_agent(
#         agentName="testname1",
#         agentResourceRoleArn="test-agent-arn",
#         tags={
#             "Key": "test-tag-key",
#         },
#     )
#     client.create_agent(
#         agentName="testname2",
#         agentResourceRoleArn="test-agent-arn",
#         tags={
#             "Key": "test-tag-key",
#         },
#     )
#     resp = client.list_agents(nextToken="1")

#     assert len(resp["agentSummaries"]) == 1
#     assert resp["agentSummaries"][0]["agentName"] == "testname2"


# @mock_aws
# def test_list_agents_bad_token():
#     client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
#     client.create_agent(
#         agentName="testname1",
#         agentResourceRoleArn="test-agent-arn",
#         tags={
#             "Key": "test-tag-key",
#         },
#     )
#     client.create_agent(
#         agentName="testname2",
#         agentResourceRoleArn="test-agent-arn",
#         tags={
#             "Key": "test-tag-key",
#         },
#     )
#     with pytest.raises(ClientError) as e:
#         client.list_agents(nextToken="3")
#     assert e.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_list_agents_max_results():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    client.create_agent(
        agentName="testname1",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )
    client.create_agent(
        agentName="testname2",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )

    resp = client.list_agents(maxResults=1)

    assert len(resp["agentSummaries"]) == 1
    assert resp["agentSummaries"][0]["agentName"] == "testname1"


@mock_aws
def test_list_agents_big_max_results():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    client.create_agent(
        agentName="testname1",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )
    client.create_agent(
        agentName="testname2",
        agentResourceRoleArn="test-agent-arn",
        tags={
            "Key": "test-tag-key",
        },
    )

    resp = client.list_agents(maxResults=4)

    assert len(resp["agentSummaries"]) == 2
    assert resp["agentSummaries"][0]["agentName"] == "testname1"
    assert resp["agentSummaries"][1]["agentName"] == "testname2"


@mock_aws
def test_delete_agent_not_found():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
        tags={"Key": "test-tag-key"},
    )
    with pytest.raises(ClientError) as e:
        client.delete_agent(
            agentId="non-existent-agent-id", skipResourceInUseCheck=True
        )
    assert e.value.response["Error"]["Code"] == "ResourceNotFoundException"


# @mock_aws
# def test_delete_agent_in_use():
#     client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
#     agent = client.create_agent(
#         agentName="testname",
#         agentResourceRoleArn="test-agent-arn",
#         tags={"Key": "test-tag-key"},
#     )
#     agent_id = agent["agent"]["agentId"]
#     with patch("moto.bedrockagent.models.Agent.agent_status", return_value="IN_USE"):
#         # mock = Mock(spec = Agent)
#         # mock.agent_status.return_value = "IN_USE"
#         with pytest.raises(ClientError) as e:
#             resp = client.delete_agent(agentId=agent_id, skipResourceInUseCheck=False)
#         assert e.value.response["Error"]["Code"] == "ConflictException"


# @mock_aws
# def test_list_knowledge_bases_token():
#     client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
#     resp = client.create_knowledge_base(
#         name="testkb",
#         description="description",
#         roleArn="test_role_arn",
#         knowledgeBaseConfiguration={
#             "type": "VECTOR",
#         },
#         storageConfiguration={
#             "type": "OPENSEARCH_SERVERLESS",
#         },
#         tags={"Key": "test-tag"},
#     )
#     resp = client.create_knowledge_base(
#         name="testkb2",
#         description="description",
#         roleArn="test_role_arn",
#         knowledgeBaseConfiguration={
#             "type": "VECTOR",
#         },
#         storageConfiguration={
#             "type": "OPENSEARCH_SERVERLESS",
#         },
#         tags={"Key": "test-tag"},
#     )
#     resp = client.list_knowledge_bases(nextToken="1")

#     assert len(resp["knowledgeBaseSummaries"]) == 1
#     assert resp["knowledgeBaseSummaries"][0]["name"] == "testkb2"


# @mock_aws
# def test_list_knowledge_bases_bad_token():
#     client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
#     client.create_knowledge_base(
#         name="testkb",
#         description="description",
#         roleArn="test_role_arn",
#         knowledgeBaseConfiguration={
#             "type": "VECTOR",
#         },
#         storageConfiguration={
#             "type": "OPENSEARCH_SERVERLESS",
#         },
#         tags={"Key": "test-tag"},
#     )
#     client.create_knowledge_base(
#         name="testkb2",
#         description="description",
#         roleArn="test_role_arn",
#         knowledgeBaseConfiguration={
#             "type": "VECTOR",
#         },
#         storageConfiguration={
#             "type": "OPENSEARCH_SERVERLESS",
#         },
#         tags={"Key": "test-tag"},
#     )
#     with pytest.raises(ClientError) as e:
#         client.list_knowledge_bases(nextToken="3")
#     assert e.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_list_knowledge_bases_max_results():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    resp = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )
    resp = client.create_knowledge_base(
        name="testkb2",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )

    resp = client.list_knowledge_bases(maxResults=1)

    assert len(resp["knowledgeBaseSummaries"]) == 1
    assert resp["knowledgeBaseSummaries"][0]["name"] == "testkb"


@mock_aws
def test_list_knowledge_bases_big_max_results():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    resp = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )
    resp = client.create_knowledge_base(
        name="testkb2",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
        tags={"Key": "test-tag"},
    )

    resp = client.list_knowledge_bases(maxResults=4)

    assert len(resp["knowledgeBaseSummaries"]) == 2
    assert resp["knowledgeBaseSummaries"][0]["name"] == "testkb"
    assert resp["knowledgeBaseSummaries"][1]["name"] == "testkb2"


@mock_aws
def test_tag_resource_knowledge_base_not_found():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    kb = client.create_knowledge_base(
        name="testkb",
        description="description",
        roleArn="test_role_arn",
        knowledgeBaseConfiguration={
            "type": "VECTOR",
        },
        storageConfiguration={
            "type": "OPENSEARCH_SERVERLESS",
        },
    )
    kb_arn = kb["knowledgeBase"]["knowledgeBaseArn"]
    with pytest.raises(ClientError) as e:
        client.tag_resource(resourceArn=kb_arn + "no", tags={"Key": "test-tag"})
    assert e.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_untag_resource_agent_not_found():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    agent = client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
    )
    agent_arn = agent["agent"]["agentArn"]
    client.tag_resource(
        resourceArn=agent_arn, tags={"Key1": "test-tag", "Key2": "test-tag2"}
    )
    with pytest.raises(ClientError) as e:
        client.untag_resource(resourceArn=agent_arn + "no", tagKeys=["Key1"])
    assert e.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_tags_for_resource_agent_not_found():
    client = boto3.client("bedrock-agent", region_name=DEFAULT_REGION)
    agent = client.create_agent(
        agentName="testname",
        agentResourceRoleArn="test-agent-arn",
    )
    agent_arn = agent["agent"]["agentArn"]
    client.tag_resource(
        resourceArn=agent_arn, tags={"Key1": "test-tag", "Key2": "test-tag2"}
    )
    with pytest.raises(ClientError) as e:
        client.list_tags_for_resource(resourceArn=agent_arn + "no")
    assert e.value.response["Error"]["Code"] == "ResourceNotFoundException"



from .responses import PanoramaResponse

url_bases = [
    r"https?://panorama\.(.+)\.amazonaws.com",
]

url_paths = {
    "{0}/$": PanoramaResponse.dispatch,
    "{0}/devices$": PanoramaResponse.dispatch,
    "{0}/devices/(?P<DeviceId>[^/]+)$": PanoramaResponse.dispatch,
    "{0}/packages/template-job$": PanoramaResponse.dispatch,
    "{0}/packages/template-job/(?P<JobId>[^/]+)$": PanoramaResponse.dispatch,
    "{0}/nodes$": PanoramaResponse.dispatch,
}



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

name = "myrule"
payload = {
    "sql": "SELECT * FROM 'topic/*' WHERE something > 0",
    "actions": [
        {"dynamoDBv2": {"putItem": {"tableName": "my-table"}, "roleArn": "my-role"}}
    ],
    "errorAction": {
        "republish": {"qos": 0, "roleArn": "my-role", "topic": "other-topic"}
    },
    "description": "my-description",
    "ruleDisabled": False,
    "awsIotSqlVersion": "2016-03-23",
}


@mock_aws
@pytest.mark.parametrize("invalid_name", ["t-r", "t.r", "t r"])
def test_topic_rule_create_with_invalid_name(invalid_name):
    client = boto3.client("iot", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.create_topic_rule(ruleName=invalid_name, topicRulePayload=payload)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidRequestException"
    assert "Member must satisfy regular expression pattern" in err["Message"]


@mock_aws
def test_topic_rule_create():
    client = boto3.client("iot", region_name="ap-northeast-1")

    client.create_topic_rule(ruleName=name, topicRulePayload=payload)

    # duplicated rule name
    with pytest.raises(ClientError) as ex:
        client.create_topic_rule(ruleName=name, topicRulePayload=payload)
    error_code = ex.value.response["Error"]["Code"]
    assert error_code == "ResourceAlreadyExistsException"


@mock_aws
def test_topic_rule_list():
    client = boto3.client("iot", region_name="ap-northeast-1")

    # empty response
    res = client.list_topic_rules()
    assert len(res["rules"]) == 0

    client.create_topic_rule(ruleName=name, topicRulePayload=payload)
    client.create_topic_rule(ruleName=f"{name}2", topicRulePayload=payload)

    res = client.list_topic_rules()
    assert len(res["rules"]) == 2
    for rule, rule_name in zip(res["rules"], [name, f"{name}2"]):
        assert rule["ruleName"] == rule_name
        assert rule["createdAt"] is not None
        assert rule["ruleArn"] is not None
        assert rule["ruleDisabled"] == payload["ruleDisabled"]
        assert rule["topicPattern"] == "topic/*"


@mock_aws
def test_topic_rule_get():
    client = boto3.client("iot", region_name="ap-northeast-1")

    # no such rule
    with pytest.raises(ClientError) as ex:
        client.get_topic_rule(ruleName=name)
    error_code = ex.value.response["Error"]["Code"]
    assert error_code == "ResourceNotFoundException"

    client.create_topic_rule(ruleName=name, topicRulePayload=payload)

    rule = client.get_topic_rule(ruleName=name)

    assert rule["ruleArn"] is not None
    rrule = rule["rule"]
    assert rrule["actions"] == payload["actions"]
    assert rrule["awsIotSqlVersion"] == payload["awsIotSqlVersion"]
    assert rrule["createdAt"] is not None
    assert rrule["description"] == payload["description"]
    assert rrule["errorAction"] == payload["errorAction"]
    assert rrule["ruleDisabled"] == payload["ruleDisabled"]
    assert rrule["ruleName"] == name
    assert rrule["sql"] == payload["sql"]


@mock_aws
def test_topic_rule_replace():
    client = boto3.client("iot", region_name="ap-northeast-1")

    # no such rule
    with pytest.raises(ClientError) as ex:
        client.replace_topic_rule(ruleName=name, topicRulePayload=payload)
    error_code = ex.value.response["Error"]["Code"]
    assert error_code == "ResourceNotFoundException"

    client.create_topic_rule(ruleName=name, topicRulePayload=payload)

    my_payload = payload.copy()
    my_payload["description"] = "new-description"
    client.replace_topic_rule(ruleName=name, topicRulePayload=my_payload)

    rule = client.get_topic_rule(ruleName=name)
    assert rule["rule"]["ruleName"] == name
    assert rule["rule"]["description"] == my_payload["description"]


@mock_aws
def test_topic_rule_disable():
    client = boto3.client("iot", region_name="ap-northeast-1")

    # no such rule
    with pytest.raises(ClientError) as ex:
        client.disable_topic_rule(ruleName=name)
    error_code = ex.value.response["Error"]["Code"]
    assert error_code == "ResourceNotFoundException"

    client.create_topic_rule(ruleName=name, topicRulePayload=payload)

    client.disable_topic_rule(ruleName=name)

    rule = client.get_topic_rule(ruleName=name)
    assert rule["rule"]["ruleName"] == name
    assert rule["rule"]["ruleDisabled"] is True


@mock_aws
def test_topic_rule_enable():
    client = boto3.client("iot", region_name="ap-northeast-1")

    # no such rule
    with pytest.raises(ClientError) as ex:
        client.enable_topic_rule(ruleName=name)
    error_code = ex.value.response["Error"]["Code"]
    assert error_code == "ResourceNotFoundException"

    my_payload = payload.copy()
    my_payload["ruleDisabled"] = True
    client.create_topic_rule(ruleName=name, topicRulePayload=my_payload)

    client.enable_topic_rule(ruleName=name)

    rule = client.get_topic_rule(ruleName=name)
    assert rule["rule"]["ruleName"] == name
    assert rule["rule"]["ruleDisabled"] is False


@mock_aws
def test_topic_rule_delete():
    client = boto3.client("iot", region_name="ap-northeast-1")

    # no such rule
    with pytest.raises(ClientError) as ex:
        client.delete_topic_rule(ruleName=name)
    error_code = ex.value.response["Error"]["Code"]
    assert error_code == "ResourceNotFoundException"

    client.create_topic_rule(ruleName=name, topicRulePayload=payload)

    client.enable_topic_rule(ruleName=name)

    client.delete_topic_rule(ruleName=name)

    res = client.list_topic_rules()
    assert len(res["rules"]) == 0



"""Unit tests for personalize-supported APIs."""

import json
import re

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_schema():
    client = boto3.client("personalize", region_name="ap-southeast-1")
    schema = {
        "type": "record",
        "name": "Interactions",
        "namespace": "com.amazonaws.personalize.schema",
        "fields": [
            {"name": "USER_ID", "type": "string"},
            {"name": "ITEM_ID", "type": "string"},
            {"name": "TIMESTAMP", "type": "long"},
        ],
        "version": "1.0",
    }

    create_schema_response = client.create_schema(
        name="personalize-demo-schema", schema=json.dumps(schema)
    )
    assert create_schema_response["schemaArn"] == (
        f"arn:aws:personalize:ap-southeast-1:{DEFAULT_ACCOUNT_ID}"
        ":schema/personalize-demo-schema"
    )


@mock_aws
def test_delete_schema():
    client = boto3.client("personalize", region_name="us-east-1")
    schema_arn = client.create_schema(name="myname", schema=json.dumps("sth"))[
        "schemaArn"
    ]
    client.delete_schema(schemaArn=schema_arn)

    assert client.list_schemas()["schemas"] == []


@mock_aws
def test_delete_schema__unknown():
    arn = f"arn:aws:personalize:ap-southeast-1:{DEFAULT_ACCOUNT_ID}:schema/personalize-demo-schema"
    client = boto3.client("personalize", region_name="us-east-2")
    with pytest.raises(ClientError) as exc:
        client.delete_schema(schemaArn=arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == f"Resource Arn {arn} does not exist."


@mock_aws
def test_describe_schema():
    client = boto3.client("personalize", region_name="us-east-2")
    schema_arn = client.create_schema(name="myname", schema="sth")["schemaArn"]
    resp = client.describe_schema(schemaArn=schema_arn)
    assert "schema" in resp
    schema = resp["schema"]

    assert schema["name"] == "myname"
    assert re.search("schema/myname", schema["schemaArn"])
    assert schema["schema"] == "sth"
    assert "creationDateTime" in schema
    assert "lastUpdatedDateTime" in schema


@mock_aws
def test_describe_schema__with_domain():
    client = boto3.client("personalize", region_name="us-east-2")
    schema_arn = client.create_schema(name="myname", schema="sth", domain="ECOMMERCE")[
        "schemaArn"
    ]
    resp = client.describe_schema(schemaArn=schema_arn)
    assert "schema" in resp
    schema = resp["schema"]

    assert schema["domain"] == "ECOMMERCE"


@mock_aws
def test_describe_schema__unknown():
    arn = (
        "arn:aws:personalize:ap-southeast-1:486285699788:schema/personalize-demo-schema"
    )
    client = boto3.client("personalize", region_name="us-east-2")
    with pytest.raises(ClientError) as exc:
        client.describe_schema(schemaArn=arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == f"Resource Arn {arn} does not exist."


@mock_aws
def test_list_schemas__initial():
    client = boto3.client("personalize", region_name="us-east-2")
    resp = client.list_schemas()

    assert resp["schemas"] == []


@mock_aws
def test_list_schema():
    client = boto3.client("personalize", region_name="us-east-2")
    schema_arn = client.create_schema(name="myname", schema="sth")["schemaArn"]

    resp = client.list_schemas()
    assert len(resp["schemas"]) == 1
    schema = resp["schemas"][0]

    assert schema["name"] == "myname"
    assert schema["schemaArn"] == schema_arn
    assert "schema" not in schema
    assert "creationDateTime" in schema
    assert "lastUpdatedDateTime" in schema



import json
from typing import Dict, List
from urllib.parse import unquote

from moto.core.responses import TYPE_RESPONSE, BaseResponse
from moto.utilities.utils import merge_multiple_dicts

from .exceptions import InvalidRequestInput
from .models import APIGatewayBackend, apigateway_backends
from .utils import deserialize_body

API_KEY_SOURCES = ["AUTHORIZER", "HEADER"]
AUTHORIZER_TYPES = ["TOKEN", "REQUEST", "COGNITO_USER_POOLS"]
ENDPOINT_CONFIGURATION_TYPES = ["PRIVATE", "EDGE", "REGIONAL"]


class APIGatewayResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="apigateway")

    def error(self, type_: str, message: str, status: int = 400) -> TYPE_RESPONSE:
        headers = self.response_headers or {}
        headers["status"] = f"{status}"
        headers["X-Amzn-Errortype"] = type_
        return (status, headers, json.dumps({"__type": type_, "message": message}))

    @property
    def backend(self) -> APIGatewayBackend:
        return apigateway_backends[self.current_account][self.region]

    def __validate_api_key_source(self, api_key_source: str) -> TYPE_RESPONSE:  # type: ignore[return]
        if api_key_source and api_key_source not in API_KEY_SOURCES:
            return self.error(
                "ValidationException",
                (
                    "1 validation error detected: "
                    "Value '{api_key_source}' at 'createRestApiInput.apiKeySource' failed "
                    "to satisfy constraint: Member must satisfy enum value set: "
                    "[AUTHORIZER, HEADER]"
                ).format(api_key_source=api_key_source),
            )

    def __validate_endpoint_configuration(  # type: ignore[return]
        self, endpoint_configuration: Dict[str, str]
    ) -> TYPE_RESPONSE:
        if endpoint_configuration and "types" in endpoint_configuration:
            invalid_types = list(
                set(endpoint_configuration["types"]) - set(ENDPOINT_CONFIGURATION_TYPES)
            )
            if invalid_types:
                return self.error(
                    "ValidationException",
                    (
                        "1 validation error detected: Value '{endpoint_type}' "
                        "at 'createRestApiInput.endpointConfiguration.types' failed "
                        "to satisfy constraint: Member must satisfy enum value set: "
                        "[PRIVATE, EDGE, REGIONAL]"
                    ).format(endpoint_type=invalid_types[0]),
                )

    def create_rest_api(self) -> TYPE_RESPONSE:
        api_doc = deserialize_body(self.body)
        if api_doc:
            fail_on_warnings = self._get_bool_param("failonwarnings") or False
            rest_api = self.backend.import_rest_api(api_doc, fail_on_warnings)

            return 200, {}, json.dumps(rest_api.to_dict())

        name = self._get_param("name")
        description = self._get_param("description")

        api_key_source = self._get_param("apiKeySource")
        endpoint_configuration = self._get_param("endpointConfiguration")
        tags = self._get_param("tags")
        policy = self._get_param("policy")
        minimum_compression_size = self._get_param("minimumCompressionSize")
        disable_execute_api_endpoint = self._get_param("disableExecuteApiEndpoint")

        # Param validation
        response = self.__validate_api_key_source(api_key_source)
        if response is not None:
            return response

        response = self.__validate_endpoint_configuration(endpoint_configuration)
        if response is not None:
            return response

        rest_api = self.backend.create_rest_api(
            name,
            description,
            api_key_source=api_key_source,
            endpoint_configuration=endpoint_configuration,
            tags=tags,
            policy=policy,
            minimum_compression_size=minimum_compression_size,
            disable_execute_api_endpoint=disable_execute_api_endpoint,
        )

        return 200, {}, json.dumps(rest_api.to_dict())

    def get_rest_apis(self) -> str:
        apis = self.backend.list_apis()
        return json.dumps({"item": [api.to_dict() for api in apis]})

    def __validte_rest_patch_operations(  # type: ignore[return]
        self, patch_operations: List[Dict[str, str]]
    ) -> TYPE_RESPONSE:
        for op in patch_operations:
            path = op["path"]
            if "apiKeySource" in path:
                value = op["value"]
                return self.__validate_api_key_source(value)

    def delete_rest_api(self) -> TYPE_RESPONSE:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        rest_api = self.backend.delete_rest_api(function_id)
        return 200, {}, json.dumps(rest_api.to_dict())

    def get_rest_api(self) -> TYPE_RESPONSE:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        rest_api = self.backend.get_rest_api(function_id)
        return 200, {}, json.dumps(rest_api.to_dict())

    def put_rest_api(self) -> TYPE_RESPONSE:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        mode = self._get_param("mode", "merge")
        fail_on_warnings = self._get_bool_param("failonwarnings") or False

        api_doc = deserialize_body(self.body)
        rest_api = self.backend.put_rest_api(
            function_id, api_doc, mode=mode, fail_on_warnings=fail_on_warnings
        )
        return 200, {}, json.dumps(rest_api.to_dict())

    def update_rest_api(self) -> TYPE_RESPONSE:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        patch_operations = self._get_param("patchOperations")
        response = self.__validte_rest_patch_operations(patch_operations)
        if response is not None:
            return response

        rest_api = self.backend.update_rest_api(function_id, patch_operations)
        return 200, {}, json.dumps(rest_api.to_dict())

    def get_resources(self) -> str:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        resources = self.backend.get_resources(function_id)
        return json.dumps({"item": [resource.to_dict() for resource in resources]})

    def create_resource(self) -> TYPE_RESPONSE:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        resource_id = self.path.split("/")[-1]
        path_part = self._get_param("pathPart")
        resource = self.backend.create_resource(function_id, resource_id, path_part)
        return 201, {"status": 201}, json.dumps(resource.to_dict())

    def delete_resource(self) -> TYPE_RESPONSE:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        resource_id = self.path.split("/")[-1]
        resource = self.backend.delete_resource(function_id, resource_id)
        return 202, {"status": 202}, json.dumps(resource.to_dict())

    def get_resource(self) -> str:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        resource_id = self.path.split("/")[-1]
        resource = self.backend.get_resource(function_id, resource_id)
        return json.dumps(resource.to_dict())

    def delete_method(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        self.backend.delete_method(function_id, resource_id, method_type)
        return 204, {"status": 204}, ""

    def get_method(self) -> str:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        method = self.backend.get_method(function_id, resource_id, method_type)
        return json.dumps(method.to_json())

    def put_method(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        authorization_type = self._get_param("authorizationType")
        api_key_required = self._get_param("apiKeyRequired")
        request_models = self._get_param("requestModels")
        operation_name = self._get_param("operationName")
        authorizer_id = self._get_param("authorizerId")
        authorization_scopes = self._get_param("authorizationScopes")
        request_validator_id = self._get_param("requestValidatorId")
        request_parameters = self._get_param("requestParameters")
        method = self.backend.put_method(
            function_id,
            resource_id,
            method_type,
            authorization_type,
            api_key_required,
            request_models=request_models,
            request_parameters=request_parameters,
            operation_name=operation_name,
            authorizer_id=authorizer_id,
            authorization_scopes=authorization_scopes,
            request_validator_id=request_validator_id,
        )
        return 201, {"status": 201}, json.dumps(method.to_json())

    def delete_method_response(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        response_code = url_path_parts[8]
        method_response = self.backend.delete_method_response(
            function_id, resource_id, method_type, response_code
        )
        return 204, {"status": 204}, json.dumps(method_response.to_json())  # type: ignore[union-attr]

    def get_method_response(self) -> str:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        response_code = url_path_parts[8]

        method_response = self.backend.get_method_response(
            function_id, resource_id, method_type, response_code
        )
        return json.dumps(method_response.to_json())  # type: ignore[union-attr]

    def put_method_response(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        response_code = url_path_parts[8]
        response_models = self._get_param("responseModels")
        response_parameters = self._get_param("responseParameters")
        method_response = self.backend.put_method_response(
            function_id,
            resource_id,
            method_type,
            response_code,
            response_models,
            response_parameters,
        )
        return 201, {"status": 201}, json.dumps(method_response.to_json())

    def create_authorizer(self) -> TYPE_RESPONSE:
        restapi_id = self.path.split("/")[2]
        name = self._get_param("name")
        authorizer_type = self._get_param("type")

        provider_arns = self._get_param("providerARNs")
        auth_type = self._get_param("authType")
        authorizer_uri = self._get_param("authorizerUri")
        authorizer_credentials = self._get_param("authorizerCredentials")
        identity_source = self._get_param("identitySource")
        identiy_validation_expression = self._get_param("identityValidationExpression")
        authorizer_result_ttl = self._get_param(
            "authorizerResultTtlInSeconds", if_none=300
        )

        # Param validation
        if authorizer_type and authorizer_type not in AUTHORIZER_TYPES:
            return self.error(
                "ValidationException",
                (
                    "1 validation error detected: "
                    "Value '{authorizer_type}' at 'createAuthorizerInput.type' failed "
                    "to satisfy constraint: Member must satisfy enum value set: "
                    "[TOKEN, REQUEST, COGNITO_USER_POOLS]"
                ).format(authorizer_type=authorizer_type),
            )

        authorizer_response = self.backend.create_authorizer(
            restapi_id=restapi_id,
            name=name,
            authorizer_type=authorizer_type,
            provider_arns=provider_arns,
            auth_type=auth_type,
            authorizer_uri=authorizer_uri,
            authorizer_credentials=authorizer_credentials,
            identity_source=identity_source,
            identiy_validation_expression=identiy_validation_expression,
            authorizer_result_ttl=authorizer_result_ttl,
        )
        return 201, {"status": 201}, json.dumps(authorizer_response.to_json())

    def delete_authorizer(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        restapi_id = url_path_parts[2]
        authorizer_id = url_path_parts[4]
        self.backend.delete_authorizer(restapi_id, authorizer_id)
        return 202, {"status": 202}, "{}"

    def get_authorizer(self) -> str:
        url_path_parts = self.path.split("/")
        restapi_id = url_path_parts[2]
        authorizer_id = url_path_parts[4]
        authorizer_response = self.backend.get_authorizer(restapi_id, authorizer_id)
        return json.dumps(authorizer_response.to_json())

    def get_authorizers(self) -> str:
        restapi_id = self.path.split("/")[2]
        authorizers = self.backend.get_authorizers(restapi_id)
        return json.dumps({"item": [a.to_json() for a in authorizers]})

    def update_authorizer(self) -> str:
        url_path_parts = self.path.split("/")
        restapi_id = url_path_parts[2]
        authorizer_id = url_path_parts[4]
        patch_operations = self._get_param("patchOperations")
        authorizer_response = self.backend.update_authorizer(
            restapi_id, authorizer_id, patch_operations
        )
        return json.dumps(authorizer_response.to_json())

    def create_request_validator(self) -> TYPE_RESPONSE:
        restapi_id = self.path.split("/")[2]
        name = self._get_param("name")
        body = self._get_bool_param("validateRequestBody")
        params = self._get_bool_param("validateRequestParameters")
        validator = self.backend.create_request_validator(
            restapi_id, name, body, params
        )
        return 201, {"status": 201}, json.dumps(validator.to_dict())

    def delete_request_validator(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        restapi_id = url_path_parts[2]
        validator_id = url_path_parts[4]
        self.backend.delete_request_validator(restapi_id, validator_id)
        return 202, {"status": 202}, ""

    def get_request_validator(self) -> str:
        url_path_parts = self.path.split("/")
        restapi_id = url_path_parts[2]
        validator_id = url_path_parts[4]
        validator = self.backend.get_request_validator(restapi_id, validator_id)
        return json.dumps(validator.to_dict())

    def get_request_validators(self) -> str:
        restapi_id = self.path.split("/")[2]
        validators = self.backend.get_request_validators(restapi_id)
        return json.dumps({"item": [validator.to_dict() for validator in validators]})

    def update_request_validator(self) -> str:
        url_path_parts = self.path.split("/")
        restapi_id = url_path_parts[2]
        validator_id = url_path_parts[4]
        patch_ops = self._get_param("patchOperations")
        validator = self.backend.update_request_validator(
            restapi_id, validator_id, patch_ops
        )
        return json.dumps(validator.to_dict())

    def create_stage(self) -> TYPE_RESPONSE:
        function_id = self.path.split("/")[2]
        stage_name = self._get_param("stageName")
        deployment_id = self._get_param("deploymentId")
        stage_variables = self._get_param("variables", if_none={})
        description = self._get_param("description", if_none="")
        cacheClusterEnabled = self._get_param("cacheClusterEnabled", if_none=False)
        cacheClusterSize = self._get_param("cacheClusterSize")
        tags = self._get_param("tags")
        tracing_enabled = self._get_param("tracingEnabled")

        stage_response = self.backend.create_stage(
            function_id,
            stage_name,
            deployment_id,
            variables=stage_variables,
            description=description,
            cacheClusterEnabled=cacheClusterEnabled,
            cacheClusterSize=cacheClusterSize,
            tags=tags,
            tracing_enabled=tracing_enabled,
        )
        return 201, {"status": 201}, json.dumps(stage_response.to_json())

    def delete_stage(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        stage_name = url_path_parts[4]
        self.backend.delete_stage(function_id, stage_name)
        return 202, {"status": 202}, "{}"

    def get_stage(self) -> str:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        stage_name = url_path_parts[4]
        stage_response = self.backend.get_stage(function_id, stage_name)
        return json.dumps(stage_response.to_json())

    def get_stages(self) -> str:
        function_id = self.path.split("/")[2]
        stages = self.backend.get_stages(function_id)
        return json.dumps({"item": [s.to_json() for s in stages]})

    def update_stage(self) -> str:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        stage_name = url_path_parts[4]
        patch_operations = self._get_param("patchOperations")
        stage_response = self.backend.update_stage(
            function_id, stage_name, patch_operations
        )
        return json.dumps(stage_response.to_json())

    def tag_resource(self) -> str:
        url_path_parts = unquote(self.path.split("/tags/")[1]).split("/")
        function_id = url_path_parts[-3]
        stage_name = url_path_parts[-1]
        tags = self._get_param("tags")
        if tags:
            stage = self.backend.get_stage(function_id, stage_name)
            stage.tags = merge_multiple_dicts(stage.tags or {}, tags)
        return json.dumps({"item": tags})

    def untag_resource(self) -> str:
        url_path_parts = unquote(self.path.split("/tags/")[1]).split("/")
        function_id = url_path_parts[-3]
        stage_name = url_path_parts[-1]
        stage = self.backend.get_stage(function_id, stage_name)
        for tag in (stage.tags or {}).copy():
            if tag in (self.querystring.get("tagKeys") or {}):
                stage.tags.pop(tag, None)  # type: ignore[union-attr]
        return json.dumps({"item": ""})

    def get_export(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        rest_api_id = url_path_parts[-5]
        export_type = url_path_parts[-1]

        body = self.backend.export_api(rest_api_id, export_type)

        now = body["info"]["version"]
        filename = f"swagger_{now}Z.json"
        headers = {
            "Content-Type": "application/octet-stream",
            "Content-Disposition": f'attachment; filename="{filename}"',
        }
        return 200, headers, json.dumps(body).encode("utf-8")

    def delete_integration(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        integration_response = self.backend.delete_integration(
            function_id, resource_id, method_type
        )
        return 204, {"status": 204}, json.dumps(integration_response.to_json())

    def get_integration(self) -> str:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        integration_response = self.backend.get_integration(
            function_id, resource_id, method_type
        )
        if integration_response:
            return json.dumps(integration_response.to_json())
        return "{}"

    def put_integration(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        integration_type = self._get_param("type")
        uri = self._get_param("uri")
        credentials = self._get_param("credentials")
        request_templates = self._get_param("requestTemplates")
        passthrough_behavior = self._get_param("passthroughBehavior")
        tls_config = self._get_param("tlsConfig")
        cache_namespace = self._get_param("cacheNamespace")
        timeout_in_millis = self._get_param("timeoutInMillis")
        request_parameters = self._get_param("requestParameters")
        content_handling = self._get_param("contentHandling")
        connection_type = self._get_param("connectionType")
        self.backend.get_method(function_id, resource_id, method_type)

        integration_http_method = self._get_param("httpMethod")

        integration_response = self.backend.put_integration(
            function_id,
            resource_id,
            method_type,
            integration_type,
            uri,
            credentials=credentials,
            integration_method=integration_http_method,
            request_templates=request_templates,
            passthrough_behavior=passthrough_behavior,
            tls_config=tls_config,
            cache_namespace=cache_namespace,
            timeout_in_millis=timeout_in_millis,
            request_parameters=request_parameters,
            content_handling=content_handling,
            connection_type=connection_type,
        )
        return 201, {"status": 201}, json.dumps(integration_response.to_json())

    def delete_integration_response(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        status_code = url_path_parts[9]
        integration_response = self.backend.delete_integration_response(
            function_id, resource_id, method_type, status_code
        )
        return 204, {"status": 204}, json.dumps(integration_response.to_json())

    def get_integration_response(self) -> str:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        status_code = url_path_parts[9]
        integration_response = self.backend.get_integration_response(
            function_id, resource_id, method_type, status_code
        )
        return json.dumps(integration_response.to_json())

    def put_integration_response(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]
        status_code = url_path_parts[9]
        if not self.body:
            raise InvalidRequestInput()

        selection_pattern = self._get_param("selectionPattern")
        response_templates = self._get_param("responseTemplates")
        response_parameters = self._get_param("responseParameters")
        content_handling = self._get_param("contentHandling")
        integration_response = self.backend.put_integration_response(
            function_id,
            resource_id,
            method_type,
            status_code,
            selection_pattern,
            response_templates,
            response_parameters,
            content_handling,
        )
        return 201, {"status": 201}, json.dumps(integration_response.to_json())

    def create_deployment(self) -> TYPE_RESPONSE:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        name = self._get_param("stageName")
        description = self._get_param("description")
        stage_variables = self._get_param("variables", if_none={})
        deployment = self.backend.create_deployment(
            function_id, name, description, stage_variables
        )
        return 201, {"status": 201}, json.dumps(deployment.to_json())

    def delete_deployment(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        deployment_id = url_path_parts[4]
        deployment = self.backend.delete_deployment(function_id, deployment_id)
        return 202, {"status": 202}, json.dumps(deployment.to_json())

    def get_deployment(self) -> str:
        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        deployment_id = url_path_parts[4]
        deployment = self.backend.get_deployment(function_id, deployment_id)
        return json.dumps(deployment.to_json())

    def get_deployments(self) -> str:
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        deployments = self.backend.get_deployments(function_id)
        return json.dumps({"item": [d.to_json() for d in deployments]})

    def create_api_key(self) -> TYPE_RESPONSE:
        apikey_response = self.backend.create_api_key(json.loads(self.body))
        return 201, {"status": 201}, json.dumps(apikey_response.to_json())

    def delete_api_key(self) -> TYPE_RESPONSE:
        apikey = self.path.split("/")[2]
        self.backend.delete_api_key(apikey)
        return 202, {"status": 202}, "{}"

    def get_api_key(self) -> str:
        apikey = self.path.split("/")[2]
        include_value = self._get_bool_param("includeValue") or False
        apikey_resp = self.backend.get_api_key(apikey).to_json()
        if not include_value:
            apikey_resp.pop("value")
        return json.dumps(apikey_resp)

    def get_api_keys(self) -> str:
        include_values = self._get_bool_param("includeValues") or False
        apikeys_response = self.backend.get_api_keys()
        resp = [a.to_json() for a in apikeys_response]
        if not include_values:
            for key in resp:
                key.pop("value")
        return json.dumps({"item": resp})

    def update_api_key(self) -> str:
        apikey = self.path.split("/")[2]
        patch_operations = self._get_param("patchOperations")
        apikey_resp = self.backend.update_api_key(apikey, patch_operations).to_json()
        return json.dumps(apikey_resp)

    def create_usage_plan(self) -> TYPE_RESPONSE:
        usage_plan_response = self.backend.create_usage_plan(json.loads(self.body))
        return 201, {"status": 201}, json.dumps(usage_plan_response.to_json())

    def delete_usage_plan(self) -> TYPE_RESPONSE:
        usage_plan = self.path.split("/")[2]
        self.backend.delete_usage_plan(usage_plan)
        return 202, {"status": 202}, "{}"

    def get_usage_plan(self) -> str:
        usage_plan = self.path.split("/")[2]

        usage_plan_response = self.backend.get_usage_plan(usage_plan)
        return json.dumps(usage_plan_response.to_json())

    def get_usage_plans(self) -> str:
        api_key_id = self.querystring.get("keyId", [None])[0]
        usage_plans_response = self.backend.get_usage_plans(api_key_id=api_key_id)
        return json.dumps({"item": [u.to_json() for u in usage_plans_response]})

    def update_usage_plan(self) -> str:
        usage_plan = self.path.split("/")[2]
        patch_operations = self._get_param("patchOperations")
        usage_plan_response = self.backend.update_usage_plan(
            usage_plan, patch_operations
        )
        return json.dumps(usage_plan_response.to_json())

    def create_usage_plan_key(self) -> TYPE_RESPONSE:
        usage_plan_id = self.path.split("/")[2]
        usage_plan = self.backend.create_usage_plan_key(
            usage_plan_id, json.loads(self.body)
        )
        return 201, {"status": 201}, json.dumps(usage_plan.to_json())

    def delete_usage_plan_key(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        usage_plan_id = url_path_parts[2]
        key_id = url_path_parts[4]
        self.backend.delete_usage_plan_key(usage_plan_id, key_id)
        return 202, {"status": 202}, "{}"

    def get_usage_plan_key(self) -> str:
        url_path_parts = self.path.split("/")
        usage_plan_id = url_path_parts[2]
        key_id = url_path_parts[4]
        usage_plan = self.backend.get_usage_plan_key(usage_plan_id, key_id)
        return json.dumps(usage_plan.to_json())

    def get_usage_plan_keys(self) -> str:
        usage_plan_id = self.path.split("/")[2]
        usage_plans_response = self.backend.get_usage_plan_keys(usage_plan_id)
        return json.dumps({"item": [u.to_json() for u in usage_plans_response]})

    def create_domain_name(self) -> TYPE_RESPONSE:
        domain_name = self._get_param("domainName")
        certificate_name = self._get_param("certificateName")
        tags = self._get_param("tags")
        certificate_arn = self._get_param("certificateArn")
        certificate_body = self._get_param("certificateBody")
        certificate_private_key = self._get_param("certificatePrivateKey")
        certificate_chain = self._get_param("certificateChain")
        regional_certificate_name = self._get_param("regionalCertificateName")
        regional_certificate_arn = self._get_param("regionalCertificateArn")
        endpoint_configuration = self._get_param("endpointConfiguration")
        security_policy = self._get_param("securityPolicy")
        domain_name_resp = self.backend.create_domain_name(
            domain_name,
            certificate_name,
            tags,
            certificate_arn,
            certificate_body,
            certificate_private_key,
            certificate_chain,
            regional_certificate_name,
            regional_certificate_arn,
            endpoint_configuration,
            security_policy,
        )
        return 201, {"status": 201}, json.dumps(domain_name_resp.to_json())

    def delete_domain_name(self) -> TYPE_RESPONSE:
        domain_name = self.path.split("/")[2]
        self.backend.delete_domain_name(domain_name)
        return 202, {"status": 202}, "{}"

    def get_domain_name(self) -> str:
        domain_name = self.path.split("/")[2]
        domain = self.backend.get_domain_name(domain_name)
        return json.dumps(domain.to_json())

    def get_domain_names(self) -> str:
        domain_names = self.backend.get_domain_names()
        return json.dumps({"item": [d.to_json() for d in domain_names]})

    def create_model(self) -> TYPE_RESPONSE:
        rest_api_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        name = self._get_param("name")
        description = self._get_param("description")
        schema = self._get_param("schema")
        content_type = self._get_param("contentType")
        model = self.backend.create_model(
            rest_api_id,
            name,
            content_type,
            description,
            schema,
        )
        return 201, {"status": 201}, json.dumps(model.to_json())

    def get_models(self) -> str:
        rest_api_id = self.path.replace("/restapis/", "", 1).split("/")[0]
        models = self.backend.get_models(rest_api_id)
        return json.dumps({"item": [m.to_json() for m in models]})

    def get_model(self) -> str:
        url_path_parts = self.path.split("/")
        rest_api_id = url_path_parts[2]
        model_name = url_path_parts[4]
        model_info = self.backend.get_model(rest_api_id, model_name)
        return json.dumps(model_info.to_json())

    def create_base_path_mapping(self) -> TYPE_RESPONSE:
        domain_name = self.path.split("/")[2]
        base_path = self._get_param("basePath")
        rest_api_id = self._get_param("restApiId")
        stage = self._get_param("stage")

        base_path_mapping = self.backend.create_base_path_mapping(
            domain_name, rest_api_id, base_path, stage
        )
        return 201, {"status": 201}, json.dumps(base_path_mapping.to_json())

    def delete_base_path_mapping(self) -> TYPE_RESPONSE:
        url_path_parts = self.path.split("/")
        domain_name = url_path_parts[2]
        base_path = unquote(url_path_parts[4])
        self.backend.delete_base_path_mapping(domain_name, base_path)
        return 202, {"status": 202}, ""

    def get_base_path_mapping(self) -> str:
        url_path_parts = self.path.split("/")
        domain_name = url_path_parts[2]
        base_path = unquote(url_path_parts[4])
        base_path_mapping = self.backend.get_base_path_mapping(domain_name, base_path)
        return json.dumps(base_path_mapping.to_json())

    def get_base_path_mappings(self) -> str:
        domain_name = self.path.split("/")[2]
        base_path_mappings = self.backend.get_base_path_mappings(domain_name)
        return json.dumps({"item": [m.to_json() for m in base_path_mappings]})

    def update_base_path_mapping(self) -> str:
        url_path_parts = self.path.split("/")
        domain_name = url_path_parts[2]
        base_path = unquote(url_path_parts[4])
        patch_ops = self._get_param("patchOperations")
        base_path_mapping = self.backend.update_base_path_mapping(
            domain_name, base_path, patch_ops
        )
        return json.dumps(base_path_mapping.to_json())

    def create_vpc_link(self) -> TYPE_RESPONSE:
        name = self._get_param("name")
        description = self._get_param("description")
        target_arns = self._get_param("targetArns")
        tags = self._get_param("tags")
        vpc_link = self.backend.create_vpc_link(
            name=name, description=description, target_arns=target_arns, tags=tags
        )
        return 202, {"status": 202}, json.dumps(vpc_link.to_json())

    def delete_vpc_link(self) -> TYPE_RESPONSE:
        vpc_link_id = self.path.split("/")[-1]
        self.backend.delete_vpc_link(vpc_link_id=vpc_link_id)
        return 202, {"status": 202}, "{}"

    def get_vpc_link(self) -> str:
        vpc_link_id = self.path.split("/")[-1]
        vpc_link = self.backend.get_vpc_link(vpc_link_id=vpc_link_id)
        return json.dumps(vpc_link.to_json())

    def get_vpc_links(self) -> str:
        vpc_links = self.backend.get_vpc_links()
        return json.dumps({"item": [v.to_json() for v in vpc_links]})

    def put_gateway_response(self) -> TYPE_RESPONSE:
        rest_api_id = self.path.split("/")[-3]
        response_type = self.path.split("/")[-1]
        params = json.loads(self.body)
        status_code = params.get("statusCode")
        response_parameters = params.get("responseParameters")
        response_templates = params.get("responseTemplates")
        response = self.backend.put_gateway_response(
            rest_api_id=rest_api_id,
            response_type=response_type,
            status_code=status_code,
            response_parameters=response_parameters,
            response_templates=response_templates,
        )
        return 201, {}, json.dumps(response.to_json())

    def get_gateway_response(self) -> TYPE_RESPONSE:
        rest_api_id = self.path.split("/")[-3]
        response_type = self.path.split("/")[-1]
        response = self.backend.get_gateway_response(
            rest_api_id=rest_api_id, response_type=response_type
        )
        return 200, {}, json.dumps(response.to_json())

    def get_gateway_responses(self) -> TYPE_RESPONSE:
        rest_api_id = self.path.split("/")[-2]
        responses = self.backend.get_gateway_responses(rest_api_id=rest_api_id)
        return 200, {}, json.dumps(dict(item=[gw.to_json() for gw in responses]))

    def delete_gateway_response(self) -> TYPE_RESPONSE:
        rest_api_id = self.path.split("/")[-3]
        response_type = self.path.split("/")[-1]
        self.backend.delete_gateway_response(
            rest_api_id=rest_api_id, response_type=response_type
        )
        return 202, {}, json.dumps(dict())



import json

from moto.core.responses import BaseResponse

from .models import ManagedBlockchainBackend, managedblockchain_backends
from .utils import (
    invitationid_from_managedblockchain_url,
    memberid_from_managedblockchain_request,
    networkid_from_managedblockchain_url,
    nodeid_from_managedblockchain_url,
    proposalid_from_managedblockchain_url,
)


class ManagedBlockchainResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="managedblockchain")

    @property
    def backend(self) -> ManagedBlockchainBackend:
        return managedblockchain_backends[self.current_account][self.region]

    def list_networks(self) -> str:
        networks = self.backend.list_networks()
        return json.dumps({"Networks": [network.to_dict() for network in networks]})

    def create_network(self) -> str:
        name = self._get_param("Name")
        framework = self._get_param("Framework")
        frameworkversion = self._get_param("FrameworkVersion")
        frameworkconfiguration = self._get_param("FrameworkConfiguration")
        voting_policy = self._get_param("VotingPolicy")
        member_configuration = self._get_param("MemberConfiguration")

        # Optional
        description = self._get_param("Description", None)

        response = self.backend.create_network(
            name,
            framework,
            frameworkversion,
            frameworkconfiguration,
            voting_policy,
            member_configuration,
            description,
        )
        return json.dumps(response)

    def get_network(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        mbcnetwork = self.backend.get_network(network_id)
        return json.dumps({"Network": mbcnetwork.get_format()})

    def list_proposals(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        proposals = self.backend.list_proposals(network_id)
        return json.dumps({"Proposals": [proposal.to_dict() for proposal in proposals]})

    def create_proposal(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        memberid = self._get_param("MemberId")
        actions = self._get_param("Actions")

        # Optional
        description = self._get_param("Description", None)

        response = self.backend.create_proposal(
            network_id, memberid, actions, description
        )
        return json.dumps(response)

    def get_proposal(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        proposal_id = proposalid_from_managedblockchain_url(self.path)
        proposal = self.backend.get_proposal(network_id, proposal_id)
        return json.dumps({"Proposal": proposal.get_format()})

    def list_proposal_votes(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        proposal_id = proposalid_from_managedblockchain_url(self.path)
        proposalvotes = self.backend.list_proposal_votes(network_id, proposal_id)
        return json.dumps({"ProposalVotes": proposalvotes})

    def vote_on_proposal(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        proposal_id = proposalid_from_managedblockchain_url(self.path)
        votermemberid = self._get_param("VoterMemberId")
        vote = self._get_param("Vote")

        self.backend.vote_on_proposal(network_id, proposal_id, votermemberid, vote)
        return ""

    def list_invitations(self) -> str:
        invitations = self.backend.list_invitations()
        return json.dumps(
            {"Invitations": [invitation.to_dict() for invitation in invitations]}
        )

    def reject_invitation(self) -> str:
        invitation_id = invitationid_from_managedblockchain_url(self.path)
        self.backend.reject_invitation(invitation_id)
        return ""

    def list_members(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        members = self.backend.list_members(network_id)
        return json.dumps({"Members": [member.to_dict() for member in members]})

    def create_member(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        invitationid = self._get_param("InvitationId")
        member_configuration = self._get_param("MemberConfiguration")

        response = self.backend.create_member(
            invitationid, network_id, member_configuration
        )
        return json.dumps(response)

    def get_member(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        member_id = memberid_from_managedblockchain_request(self.uri, self.body)
        member = self.backend.get_member(network_id, member_id)
        return json.dumps({"Member": member.get_format()})

    def update_member(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        member_id = memberid_from_managedblockchain_request(self.uri, self.body)
        logpublishingconfiguration = self._get_param("LogPublishingConfiguration")
        self.backend.update_member(network_id, member_id, logpublishingconfiguration)
        return ""

    def delete_member(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        member_id = memberid_from_managedblockchain_request(self.uri, self.body)
        self.backend.delete_member(network_id, member_id)
        return ""

    def list_nodes(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        member_id = memberid_from_managedblockchain_request(self.uri, self.body)
        status = self._get_param("status")
        nodes = self.backend.list_nodes(network_id, member_id, status)
        return json.dumps({"Nodes": [node.to_dict() for node in nodes]})

    def create_node(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        member_id = memberid_from_managedblockchain_request(self.uri, self.body)
        instancetype = self._get_param("NodeConfiguration")["InstanceType"]
        availabilityzone = self._get_param("NodeConfiguration")["AvailabilityZone"]
        logpublishingconfiguration = self._get_param("NodeConfiguration")[
            "LogPublishingConfiguration"
        ]

        response = self.backend.create_node(
            network_id,
            member_id,
            availabilityzone,
            instancetype,
            logpublishingconfiguration,
        )
        return json.dumps(response)

    def get_node(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        member_id = memberid_from_managedblockchain_request(self.uri, self.body)
        node_id = nodeid_from_managedblockchain_url(self.path)
        node = self.backend.get_node(network_id, member_id, node_id)
        return json.dumps({"Node": node.get_format()})

    def update_node(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        member_id = memberid_from_managedblockchain_request(self.uri, self.body)
        node_id = nodeid_from_managedblockchain_url(self.path)
        self.backend.update_node(
            network_id, member_id, node_id, logpublishingconfiguration=self.body
        )
        return ""

    def delete_node(self) -> str:
        network_id = networkid_from_managedblockchain_url(self.path)
        member_id = memberid_from_managedblockchain_request(self.uri, self.body)
        node_id = nodeid_from_managedblockchain_url(self.path)
        self.backend.delete_node(network_id, member_id, node_id)
        return ""



import os
from functools import wraps
from uuid import uuid4

import boto3

from moto import mock_aws


def dynamodb_aws_verified(
    create_table: bool = True,
    add_range: bool = False,
    numeric_range: bool = False,
    add_gsi: bool = False,
    add_gsi_range: bool = False,
    add_lsi: bool = False,
    numeric_gsi_range: bool = False,
    numeric_lsi_range: bool = False,
):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.

    This decorator will:
      - Create a table
      - Run the test and pass the table_name as an argument
      - Delete the table
    """

    def inner(func):
        @wraps(func)
        def pagination_wrapper():
            table_name = "t" + str(uuid4())[0:6]

            allow_aws_request = (
                os.environ.get("MOTO_TEST_ALLOW_AWS_REQUEST", "false").lower() == "true"
            )

            if allow_aws_request:
                if create_table:
                    print(f"Test {func} will create DDB Table {table_name}")  # noqa
                    return create_table_and_test(table_name)
                else:
                    return func()
            else:
                with mock_aws():
                    if create_table:
                        return create_table_and_test(table_name)
                    else:
                        return func()

        def create_table_and_test(table_name):
            client = boto3.client("dynamodb", region_name="us-east-1")

            range_type = "N" if numeric_range else "S"
            gsi_range_type = "N" if numeric_gsi_range else "S"
            lsi_range_type = "N" if numeric_lsi_range else "S"

            kwargs = {
                "KeySchema": [{"AttributeName": "pk", "KeyType": "HASH"}],
                "AttributeDefinitions": [{"AttributeName": "pk", "AttributeType": "S"}],
            }
            if add_range or numeric_range:
                kwargs["KeySchema"].append({"AttributeName": "sk", "KeyType": "RANGE"})
                kwargs["AttributeDefinitions"].append(
                    {"AttributeName": "sk", "AttributeType": range_type}
                )
            if add_gsi:
                kwargs["GlobalSecondaryIndexes"] = [
                    {
                        "IndexName": "test_gsi",
                        "KeySchema": [
                            {"AttributeName": "gsi_pk", "KeyType": "HASH"},
                        ],
                        "Projection": {"ProjectionType": "ALL"},
                        "ProvisionedThroughput": {
                            "ReadCapacityUnits": 1,
                            "WriteCapacityUnits": 1,
                        },
                    }
                ]
                kwargs["AttributeDefinitions"].append(
                    {"AttributeName": "gsi_pk", "AttributeType": "S"}
                )
            if add_gsi_range or numeric_gsi_range:
                kwargs["GlobalSecondaryIndexes"] = [
                    {
                        "IndexName": "test_gsi",
                        "KeySchema": [
                            {"AttributeName": "gsi_pk", "KeyType": "HASH"},
                            {"AttributeName": "gsi_sk", "KeyType": "RANGE"},
                        ],
                        "Projection": {"ProjectionType": "ALL"},
                        "ProvisionedThroughput": {
                            "ReadCapacityUnits": 1,
                            "WriteCapacityUnits": 1,
                        },
                    }
                ]
                kwargs["AttributeDefinitions"].append(
                    {"AttributeName": "gsi_pk", "AttributeType": "S"}
                )
                kwargs["AttributeDefinitions"].append(
                    {"AttributeName": "gsi_sk", "AttributeType": gsi_range_type}
                )
            if add_lsi or numeric_lsi_range:
                kwargs["LocalSecondaryIndexes"] = [
                    {
                        "IndexName": "test_lsi",
                        "KeySchema": [
                            {"AttributeName": "pk", "KeyType": "HASH"},
                            {"AttributeName": "lsi_sk", "KeyType": "RANGE"},
                        ],
                        "Projection": {"ProjectionType": "ALL"},
                    }
                ]
                kwargs["AttributeDefinitions"].append(
                    {"AttributeName": "lsi_sk", "AttributeType": lsi_range_type}
                )
            client.create_table(
                TableName=table_name,
                ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 5},
                Tags=[{"Key": "environment", "Value": "moto_tests"}],
                **kwargs,
            )
            waiter = client.get_waiter("table_exists")
            waiter.wait(TableName=table_name)
            try:
                resp = func(table_name)
            finally:
                ### CLEANUP ###
                client.delete_table(TableName=table_name)

            return resp

        return pagination_wrapper

    return inner
